"""Eneboo_mdi module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtGui, QtCore, QtXml  # type: ignore[import]

from pineboolib.core.utils import utils_base
from pineboolib.core import settings

from pineboolib.application.acls import pnaccesscontrollists

from pineboolib.fllegacy.aqsobjects import aqs
from pineboolib.fllegacy import flworkspace, flformdb
from pineboolib.application import pncore
from pineboolib import application
from pineboolib import logging
from pineboolib.interfaces import imainwindow
from pineboolib.q3widgets import qmainwindow


from typing import Any, cast, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.database import pnconnectionmanager  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class MainForm(imainwindow.IMainWindow):
    """MainForm class."""

    acl_: pnaccesscontrollists.PNAccessControlLists
    is_closing_: bool
    mdi_enable_: bool

    main_window: QtWidgets.QMainWindow
    exit_button: QtWidgets.QPushButton

    mdi_toolbuttons: List[QtWidgets.QToolButton]

    tool_box_: Any
    toogle_bars_: Any
    last_text_caption_: str
    _map_geometry_form: List[Any]

    def __init__(self) -> None:
        """Inicialize."""
        super().__init__()
        self._p_work_space = None
        self.main_widget = self
        self.is_closing_ = False
        self.mdi_toolbuttons = []
        self._dict_main_widgets = {}

        self.container_ = None
        self.tool_box_ = None
        self.toogle_bars_ = None
        self._map_geometry_form = []
        self.last_text_caption_ = ""

    def reinitScript(self):
        """Reinit script."""
        # application.PROJECT.aq_app.startTimerIdle()
        # self.tool_box = None
        # self.modules_menu = None

        if self._dict_main_widgets:
            self._dict_main_widgets = {}

    def initScript(self) -> None:
        """Inicialize main script."""

        self.createUi(utils_base.filedir("plugins/mainform/eneboo_mdi/mainform.ui"))

        self.container_ = self
        self.init()

    def db(self) -> "pnconnectionmanager.PNConnectionManager":
        """Return the dababase connection."""

        return application.PROJECT.conn_manager

    def createUi(self, ui_file: str) -> None:
        """Create UI from a file."""

        self.main_window = cast(
            QtWidgets.QMainWindow, self.db().managerModules().createUI(ui_file, None, self)
        )
        self.main_window.setObjectName("container")

    def init(self) -> None:
        """Initialize FLApplication."""

        self._dict_main_widgets = {}

        if self.container_ is None:
            raise Exception("init. self.container_ is empty")

        self.container_.setObjectName("container")
        self.container_.setWindowIcon(QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("pineboo.png")))
        if self.db() is not None:
            self.container_.setWindowTitle(self.db().mainConn().DBName())
        else:
            self.container_.setWindowTitle("Pineboo %s" % application.PROJECT.load_version())

        # FLDiskCache.init(self)

        self.window_menu = QtWidgets.QMenu(self.container_)
        self.window_menu.setObjectName("windowMenu")

        self.window_cascade_action = QtGui.QAction(
            QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("cascada.png")),
            self.tr("Cascada"),
            self.container_,
        )
        self.window_menu.addAction(self.window_cascade_action)

        self.window_tile_action = QtGui.QAction(
            QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("mosaico.png")),
            self.tr("Mosaico"),
            self.container_,
        )
        self.window_menu.addAction(self.window_tile_action)

        self.window_close_action = QtGui.QAction(
            QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("cerrar.png")),
            self.tr("Cerrar"),
            self.container_,
        )
        self.window_menu.addAction(self.window_close_action)

        self.modules_menu = QtWidgets.QMenu(self.container_)
        self.modules_menu.setObjectName("modulesMenu")
        # self.modules_menu.setCheckable(False)

        widget = QtWidgets.QWidget(self.container_)
        widget.setObjectName("widgetContainer")
        vbox_layout = QtWidgets.QVBoxLayout(widget)

        self.exit_button = QtWidgets.QPushButton(
            QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("exit.png")), self.tr("Salir"), widget
        )
        self.exit_button.setObjectName("pbSalir")
        self.exit_button.setShortcut(QtGui.QKeySequence(self.tr("Ctrl+Q")))
        self.exit_button.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
            )
        )
        self.exit_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.exit_button.setToolTip(self.tr("Salir de la aplicación (Ctrl+Q)"))
        self.exit_button.setWhatsThis(self.tr("Salir de la aplicación (Ctrl+Q)"))
        self.exit_button.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
            self.exit_button_clicked
        )

        self.tool_box_ = QtWidgets.QToolBox(widget)
        self.tool_box_.setObjectName("toolBox")

        vbox_layout.addWidget(self.exit_button)
        vbox_layout.addWidget(self.tool_box_)
        self.container_.setCentralWidget(widget)

        self.db().manager().init()

        application.PROJECT.aq_app.initStyles()
        self.initMenuBar()

        self.db().manager().loadTables()
        self.db().managerModules().loadKeyFiles()
        self.db().managerModules().loadAllIdModules()
        self.db().managerModules().loadIdAreas()

        # self.acl_ = pnaccesscontrollists.PNAccessControlLists()
        # self.acl_.init()

        # self.loadScripts()
        self.db().managerModules().setShaLocalFromGlobal()
        # application.PROJECT.aq_app.loadTranslations()

        application.PROJECT.call("sys.init", [])
        self.initToolBox()
        self.readState()

        self.container_.installEventFilter(self)
        application.PROJECT.aq_app.startTimerIdle()

    def exit_button_clicked(self):
        """Event when exit button is clicked. Closes the container."""
        if self.generalExit(True):
            self.is_closing_ = True
            self.close()

    def initMenuBar(self) -> None:
        """Initialize menus."""
        if self.window_menu is None:
            raise Exception("initMenuBar. self.window_menu is empty!")

        cast(
            QtCore.pyqtSignal, self.window_menu.aboutToShow
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.windowMenuAboutToShow
        )

    def initToolBar(self) -> None:
        """Initialize toolbar."""

        if self.main_widget is None:
            return

        menu_bar = self.main_widget.menuBar()  # type: ignore [attr-defined] # noqa: F821
        if menu_bar is None:
            LOGGER.warning("No se encuentra toolbar en %s", self.main_widget.objectName())
            return

        # tb.setMovingEnabled(False)

        menu_bar.addSeparator()
        # what_this_button = QWhatsThis(tb)

        if not self.toogle_bars_:
            self.toogle_bars_ = QtWidgets.QMenu(self.container_)
            self.toogle_bars_.setObjectName("toggleBars")
            # self.toogle_bars_.setCheckable(True)

            # ag = QtGui.QActionGroup(self.container_)
            # ag.setObjectName("agToggleBars")

            tools_action = QtGui.QAction(self.tr("Barra de Herramientas"), self.container_)
            tools_action.setObjectName("Herramientas")
            tools_action.setCheckable(True)
            tools_action.setChecked(True)
            tools_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                self.toggleToolBar
            )
            self.toogle_bars_.addAction(tools_action)

            status_action = QtGui.QAction(self.tr("Barra de Estado"), self.container_)
            status_action.setObjectName("Estado")
            status_action.setCheckable(True)
            status_action.setChecked(True)
            status_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                self.toggleStatusBar
            )
            self.toogle_bars_.addAction(status_action)

            self.main_widget.menuBar().addMenu(  # type: ignore [attr-defined] # noqa: F821
                self.toogle_bars_
            )

        view_action = self.main_widget.menuBar().addMenu(  # type: ignore [attr-defined] # noqa: F821
            self.toogle_bars_
        )
        view_action.setText(self.tr("&Ver"))

        modules_action = self.main_widget.menuBar().addMenu(  # type: ignore [attr-defined] # noqa: F821
            self.modules_menu
        )
        modules_action.setText(self.tr("&Módulos"))

    def windowMenuAboutToShow(self) -> None:
        """Signal called before window menu is shown."""
        if not self._p_work_space:
            return
        if self.window_menu is None:
            return
        self.window_menu.clear()
        self.window_menu.addAction(self.window_cascade_action)
        self.window_menu.addAction(self.window_tile_action)
        self.window_menu.addAction(self.window_close_action)

        state = True if self._p_work_space.subWindowList() else False
        self.window_cascade_action.setEnabled(state)
        self.window_tile_action.setEnabled(state)
        self.window_close_action.setEnabled(state)
        if state:
            self.window_menu.addSeparator()

        for window in self._p_work_space.subWindowList():
            action = self.window_menu.addAction(window.windowTitle())
            action.setCheckable(True)

            if window == self._p_work_space.activeSubWindow():
                action.setChecked(True)

            action.triggered.connect(window.setFocus)

    def windowMenuActivated(self, id) -> None:
        """Signal called when user clicks on menu."""
        if not self._p_work_space:
            return

        window_list_ = self._p_work_space.subWindowList()
        if window_list_:
            window_list_[id].setFocus()

    def existFormInMDI(self, form_name: str) -> bool:
        """Return if named FLFormDB is open."""
        if form_name is None or not self._p_work_space:
            return False

        for window in self._p_work_space.subWindowList():
            if window.findChild(flformdb.FLFormDB).idMDI() == form_name:
                window.showNormal()
                window.setFocus()
                return True

        return False

    def windowClose(self) -> None:
        """Signal called on close."""
        if self._p_work_space is None:
            return

        self._p_work_space.closeActiveSubWindow()

    def workspace(self) -> Any:
        """Get current workspace."""
        return self._p_work_space

    def initToolBox(self) -> None:
        """Initialize toolbox."""

        self.tool_box_ = cast(QtWidgets.QToolBox, self.findChild(QtWidgets.QToolBox, "toolBox"))
        self.modules_menu = cast(QtWidgets.QMenu, self.findChild(QtWidgets.QMenu, "modulesMenu"))

        if self.tool_box_ is None or self.modules_menu is None:
            return

        self.modules_menu.clear()
        for number in reversed(range(self.tool_box_.count())):
            item = self.tool_box_.widget(number)
            if isinstance(item, QtWidgets.QToolBar):
                item.clear()

            self.tool_box_.removeItem(number)

        for button in self.mdi_toolbuttons:
            self.mdi_toolbuttons.remove(button)

        del self.mdi_toolbuttons
        self.mdi_toolbuttons = []

        char_num = 65

        for id_area in self.db().managerModules().listIdAreas():
            descript_area = self.db().managerModules().idAreaToDescription(id_area)
            new_area_bar = QtWidgets.QToolBar(self.tr(descript_area), self.container_)
            new_area_bar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
            # new_area_bar.setFrameStyle(QFrame.NoFrame)
            new_area_bar.setOrientation(QtCore.Qt.Orientation.Vertical)
            new_area_bar.layout().setSpacing(3)  # type: ignore [union-attr]
            self.tool_box_.addItem(new_area_bar, self.tr(descript_area))
            action_group = QtGui.QActionGroup(new_area_bar)
            action_group.setObjectName(descript_area)
            # ac = QtGui.QAction(ag)
            # ac.setText(descript_area)
            # ac.setUsesDropDown(True)

            list_modules = self.db().managerModules().listIdModules(id_area)
            list_modules.sort()

            for mod in list_modules:
                if str(chr(char_num)) == "Q":
                    char_num += 1
                    continue

                if mod == "sys":
                    if settings.CONFIG.value("application/isDebuggerMode", False):
                        descript_module = "%s: %s" % (
                            str(chr(char_num)),
                            self.tr("Carga Estática desde Disco Duro"),
                        )
                        new_module_action = QtGui.QAction(new_area_bar)
                        new_module_action.setObjectName("StaticLoadAction")
                        new_module_action.setText(self.tr(descript_module))
                        new_module_action.setShortcut(
                            getattr(QtCore.Qt.Key, "Key_%s" % str(chr(char_num))).value
                        )
                        new_module_action.setIcon(
                            QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("folder_update.png"))
                        )
                        new_area_bar.addAction(new_module_action)
                        new_module_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                            application.PROJECT.aq_app.staticLoaderSetup
                        )
                        action_group.addAction(new_module_action)
                        char_num += 1

                        descript_module = "%s: %s" % (
                            str(chr(char_num)),
                            self.tr("Reiniciar Script"),
                        )
                        new_module_action = QtGui.QAction(new_area_bar)
                        new_module_action.setObjectName("reinitScriptAction")
                        new_module_action.setText(self.tr(descript_module))
                        new_module_action.setShortcut(
                            getattr(QtCore.Qt.Key, "Key_%s" % str(chr(char_num))).value
                        )
                        new_module_action.setIcon(
                            QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("reload.png"))
                        )
                        new_area_bar.addAction(new_module_action)
                        new_module_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                            application.PROJECT.aq_app.reinit
                        )
                        action_group.addAction(new_module_action)
                        char_num += 1

                    descript_module = "%s: %s" % (
                        str(chr(char_num)),
                        self.tr("Mostrar Consola de mensajes"),
                    )
                    new_module_action = QtGui.QAction(new_area_bar)
                    new_module_action.setObjectName("shConsoleAction")
                    new_module_action.setText(self.tr(descript_module))
                    new_module_action.setShortcut(
                        getattr(QtCore.Qt.Key, "Key_%s" % str(chr(char_num))).value
                    )
                    new_module_action.setIcon(
                        QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("consola.png"))
                    )
                    new_area_bar.addAction(new_module_action)
                    new_module_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                        application.PROJECT.aq_app.showConsole
                    )
                    action_group.addAction(new_module_action)
                    char_num += 1

                descript_module = "%s: %s" % (
                    str(chr(char_num)),
                    self.db().managerModules().idModuleToDescription(mod),
                )
                new_module_action = QtGui.QAction(new_area_bar)
                new_module_action.setObjectName(mod)
                new_module_action.setText(self.tr(descript_module))
                new_module_action.setShortcut(
                    getattr(QtCore.Qt.Key, "Key_%s" % str(chr(char_num))).value
                )
                new_module_action.setIcon(QtGui.QIcon(self.db().managerModules().iconModule(mod)))
                new_area_bar.addAction(new_module_action)
                new_module_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.activateModule
                )
                action_group.addAction(new_module_action)
                char_num += 1

            # sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)

            lay = new_area_bar.layout()
            for child in new_area_bar.children():
                if isinstance(child, QtWidgets.QToolButton):
                    self.mdi_toolbuttons.append(child)
                    lay.setAlignment(child, QtCore.Qt.AlignmentFlag.AlignCenter)  # type: ignore [union-attr]

            a_menu = self.modules_menu.addMenu(descript_area)
            for action in action_group.actions():
                a_menu.addAction(action)  # type: ignore [union-attr]

        descript_area = "Configuración"
        config_tool_bar = QtWidgets.QToolBar(self.tr(descript_area), self.container_)
        config_tool_bar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        # config_tool_bar.setFrameStyle(QFrame.NoFrame)
        config_tool_bar.setOrientation(QtCore.Qt.Orientation.Vertical)
        # config_tool_bar.layout().setSpacing(3)
        self.tool_box_.addItem(config_tool_bar, self.tr(descript_area))

        descript_module = self.tr("Fuente")
        font_action = QtGui.QAction(config_tool_bar)
        font_action.setObjectName("fontAction")
        font_action.setText(self.tr(descript_module))
        # font_action.setShortcut(getattr(QtCore.Qt, "Key_%s" % str(chr(c))))
        font_action.setIcon(QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("font.png")))
        config_tool_bar.addAction(font_action)
        font_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.chooseFont
        )

        descript_module = self.tr("Estilo")
        style_action = QtGui.QAction(config_tool_bar)
        style_action.setObjectName("styleAction")
        style_action.setText(self.tr(descript_module))
        # style_action.setShortcut(getattr(QtCore.Qt, "Key_%s" % str(chr(c))))
        style_action.setIcon(QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("estilo.png")))
        config_tool_bar.addAction(style_action)
        style_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.showStyles
        )

        descript_module = self.tr("Indice")
        help_action = QtGui.QAction(config_tool_bar)
        help_action.setObjectName("helpAction")
        help_action.setText(self.tr(descript_module))
        # help_action.setShortcut(getattr(QtCore.Qt, "Key_%s" % str(chr(c))))
        help_action.setIcon(QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("help_index.png")))
        config_tool_bar.addAction(help_action)
        help_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.helpIndex
        )

        descript_module = self.tr("Acerca de Pineboo")
        about_pineboo_action = QtGui.QAction(config_tool_bar)
        about_pineboo_action.setObjectName("aboutPinebooAction")
        about_pineboo_action.setText(self.tr(descript_module))
        # help_action.setShortcut(getattr(QtCore.Qt, "Key_%s" % str(chr(c))))
        about_pineboo_action.setIcon(QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("about.png")))
        config_tool_bar.addAction(about_pineboo_action)
        about_pineboo_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.aboutPineboo
        )

        descript_module = self.tr("Visita Eneboo.org")
        visit_pineboo_action = QtGui.QAction(config_tool_bar)
        visit_pineboo_action.setObjectName("visitPinebooAction")
        visit_pineboo_action.setText(self.tr(descript_module))
        # help_action.setShortcut(getattr(QtCore.Qt, "Key_%s" % str(chr(c))))
        visit_pineboo_action.setIcon(QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("about.png")))
        config_tool_bar.addAction(visit_pineboo_action)
        visit_pineboo_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.urlPineboo
        )

        descript_module = self.tr("Acerca de Qt")
        about_qt_action = QtGui.QAction(config_tool_bar)
        about_qt_action.setObjectName("aboutQtAction")
        about_qt_action.setText(self.tr(descript_module))
        # help_action.setShortcut(getattr(QtCore.Qt, "Key_%s" % str(chr(c))))
        about_qt_action.setIcon(QtGui.QIcon(aqs.AQS.pixmap_fromMimeSource("aboutqt.png")))
        config_tool_bar.addAction(about_qt_action)
        about_qt_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
            application.PROJECT.aq_app.aboutQt
        )

        lay = config_tool_bar.layout()
        for child in config_tool_bar.children():
            if isinstance(child, QtWidgets.QToolButton):
                self.mdi_toolbuttons.append(child)
                lay.setAlignment(child, QtCore.Qt.AlignmentFlag.AlignCenter)  # type: ignore [union-attr]

        if application.PROJECT.aq_app.acl_:
            application.PROJECT.aq_app.acl_.process(self.container_)

    def eventFilter(
        self, obj_: Optional["QtCore.QObject"], event: Optional["QtCore.QEvent"]
    ) -> bool:
        """React to user events."""

        if self._inicializing or application.PROJECT.aq_app._destroying:
            return super().eventFilter(obj_, event)  # type: ignore [arg-type]

        # if QtWidgets.QApplication.activeModalWidget() or QtWidgets.QApplication.activePopupWidget():
        #    return super().eventFilter(obj, event)
        obj_ = cast(QtWidgets.QWidget, obj_)
        event_type = event.type()  # type: ignore [union-attr]
        main_widget = self.main_widget

        if (
            obj_ != main_widget
            and not isinstance(obj_, QtWidgets.QMainWindow)
            and not isinstance(obj_, qmainwindow.QMainWindow)
        ):
            return super().eventFilter(obj_, event)  # type: ignore [arg-type]

        # aw = None
        # if self._p_work_space is not None:
        #    aw = QtWidgets.QApplication.setActiveWindow(self._p_work_space)
        # if aw is not None and aw != obj and event_type not in (QEvent.Resize, QEvent.Close):
        #     obj.removeEventFilter(self)
        #     if event_type == QEvent.WindowActivate:
        #         if obj == self.container_:
        #             self.activateModule(None)
        #         else:
        #             self.activateModule(obj.objectName())
        #
        #     if self._p_work_space and self.notify(self._p_work_space, ev):
        #         obj.installEventFilter(self)
        #         return True
        #
        #     obj.installEventFilter(self)

        if event_type == QtCore.QEvent.Type.KeyPress:
            # if obj_ == self.container_:
            #    key_ = cast(QtGui.QKeyEvent, event)

            if obj_ == main_widget:
                key_ = cast(QtGui.QKeyEvent, event)
                if (
                    key_.key() == cast(int, QtCore.Qt.Key.Key_Shift.value)
                    and key_.modifiers()  # type: ignore [comparison-overlap] # noqa: F821
                    == QtCore.Qt.KeyboardModifier.ControlModifier  # type: ignore [comparison-overlap] # noqa: F821
                ):
                    self.activateModule(None)
                    return True
                elif (
                    key_.key() == cast(int, QtCore.Qt.Key.Key_Q.value)
                    and key_.modifiers()  # type: ignore [comparison-overlap] # noqa: F821
                    == QtCore.Qt.KeyboardModifier.ControlModifier  # type: ignore [comparison-overlap] # noqa: F821
                ):
                    self.generalExit()
                    return True
                elif key_.key() == cast(
                    int, QtCore.Qt.Key.Key_W.value
                ) and key_.modifiers() in [  # type: ignore [comparison-overlap] # noqa: F821
                    QtCore.Qt.KeyboardModifier.AltModifier,
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                ]:
                    LOGGER.warning("unknown key presset!.")
                    return True
                elif key_.key() == cast(int, QtCore.Qt.Key.Key_Escape.value):
                    obj_.hide()
                    return True

        elif event_type == QtCore.QEvent.Type.Close:
            if obj_ is self and not self.is_closing_:
                ret = self.generalExit()
                if not ret:
                    obj_.setDisabled(False)
                    event.ignore()  # type: ignore [union-attr]
            else:
                self.writeStateActions(obj_.objectName())

            return True

        elif event_type == QtCore.QEvent.Type.WindowActivate:
            if obj_ == self.container_:
                self.activateModule(None)
                return True
            else:
                self.activateModule(obj_.objectName())
                return True

        # elif event_type == QtCore.QEvent.MouseButtonPress:
        #    if self.modules_menu:
        #        me = ev
        #        if me.button() == QtCore.Qt.RightButton:
        #            self.modules_menu.pop(QtGui.QCursor.pos())
        #            return True
        #        else:
        #            return False
        #    else:
        #        return False

        elif event_type == QtCore.QEvent.Type.Resize and obj_ is self:
            for tool_button in self.mdi_toolbuttons:
                tool_button.setMinimumWidth(obj_.width() - 10)

            return True

        return super().eventFilter(obj_, event)  # type: ignore [arg-type]

    def activateModule(self, idm=None) -> None:
        """Initialize module."""

        if not idm:
            if self.sender():
                idm = self.sender().objectName()  # type: ignore [union-attr]

        if not idm:
            return

        self.writeStateModule()

        widget = None
        if idm in self.db().managerModules().listAllIdModules():
            widget = self._dict_main_widgets[idm] if idm in self._dict_main_widgets.keys() else None
            if widget is None:
                widget = self.db().managerModules().createUI(file_name="%s.ui" % idm)
                if widget is None:
                    return
                if isinstance(widget, qmainwindow.QMainWindow) or widget.findChild(pncore.PNCore):
                    doc = QtXml.QDomDocument()
                    ui_file = "%s.ui" % idm
                    content_cached = self.db().managerModules().contentCached(ui_file)
                    if not content_cached or not doc.setContent(content_cached):
                        if content_cached is None:
                            LOGGER.warning("No se ha podido cargar %s" % (ui_file))
                            return None

                    root = doc.documentElement().toElement()
                    conns = root.namedItem("connections").toElement()
                    connections = conns.elementsByTagName("connection")
                    for i in range(connections.length()):
                        itn = connections.at(i).toElement()
                        sender = itn.namedItem("sender").toElement().text()
                        signal = itn.namedItem("signal").toElement().text()

                        if signal in ["activated()", "triggered()"]:
                            signal = "triggered()"
                        receiver = itn.namedItem("receiver").toElement().text()
                        if receiver == "FLWidgetApplication":
                            receiver = idm
                        slot = itn.namedItem("slot").toElement().text()

                        if receiver in [idm, "pncore"] and signal == "triggered()":
                            action_list = []
                            action = cast(QtGui.QAction, widget.findChild(QtGui.QAction, sender))
                            if action is not None:
                                action_list.append(action)

                            for menu in widget.findChildren(QtWidgets.QToolBar):
                                action = cast(QtGui.QAction, menu.findChild(QtGui.QAction, sender))
                                if action is not None and action not in action_list:
                                    action_list.append(action)

                            for action in action_list:
                                if (
                                    action is not None
                                    and sender in application.PROJECT.actions.keys()
                                ):
                                    slot_obj = getattr(
                                        application.PROJECT.actions[sender],
                                        slot[0 : slot.find("(")],
                                    )
                                    action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                                        slot_obj
                                    )
                                else:
                                    LOGGER.warning("Action %s not found", sender)

                widget.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
                self._dict_main_widgets[idm] = widget
                widget.setObjectName(idm)
                if application.PROJECT.aq_app.acl_:
                    application.PROJECT.aq_app.acl_.process(widget)

                self.setMainWidget(widget)
                application.PROJECT.aq_app.call("%s.init()" % idm, [])
                widget.removeEventFilter(self)
                self.db().managerModules().setActiveIdModule(idm)
                self.setMainWidget(widget)
                self.initMainWidget()
                self.showMainWidget(widget)
                widget.installEventFilter(self)
                return

        self.db().managerModules().setActiveIdModule("" if widget is None else idm)

        self.setMainWidget(widget)
        self.showMainWidget(widget)

    def writeState(self) -> None:
        """Write settings back to disk."""

        settings.SETTINGS.set_value("MultiLang/Enabled", self._multi_lang_enabled)
        settings.SETTINGS.set_value("MultiLang/LangId", self._multi_lang_id)

        if self.container_ is not None:
            windows_opened = []
            _list = QtWidgets.QApplication.topLevelWidgets()

            if self._inicializing:
                for item in _list:
                    item.removeEventFilter(self)
                    if item.objectName() in self._dict_main_widgets.keys():
                        if item != self.container_:
                            if item.isVisible():
                                windows_opened.append(item.objectName())
                            item.hide()
                        else:
                            item.setDisabled(True)
            else:
                for item in _list:
                    if (
                        item != self.container_
                        and item.isVisible()
                        and item.objectName() in self._dict_main_widgets.keys()
                    ):
                        windows_opened.append(item.objectName())

            settings.SETTINGS.set_value("windowsOpened/Main", windows_opened)
            settings.SETTINGS.set_value(
                "Geometry/MainWindowMaximized", self.container_.isMaximized()
            )
            if not self.container_.isMaximized():
                settings.SETTINGS.set_value("Geometry/MainWindowX", self.container_.x())
                settings.SETTINGS.set_value("Geometry/MainWindowY", self.container_.y())
                settings.SETTINGS.set_value("Geometry/MainWindowWidth", self.container_.width())
                settings.SETTINGS.set_value("Geometry/MainWindowHeight", self.container_.height())

    def writeStateActions(self, idm: str) -> None:
        """Write state actions forms."""

        container_ = None
        for level_widget in QtWidgets.QApplication.topLevelWidgets():
            if level_widget.objectName() == idm:
                container_ = level_widget
                break

        if container_ and idm:
            for item in [
                item.widget() for item in container_.findChildren(QtWidgets.QMdiSubWindow)  # type: ignore [attr-defined]
            ]:
                if item:
                    key = "Geometry/%s/" % item._action_name  # type: ignore [attr-defined]
                    settings.SETTINGS.set_value("%s/X" % key, item.x())
                    settings.SETTINGS.set_value("%s/Y" % key, item.y())
                    settings.SETTINGS.set_value("%s/Width" % key, item.width())
                    settings.SETTINGS.set_value("%s/Height" % key, item.height())

    def writeStateModule(self) -> None:
        """Write settings for modules."""

        idm = self.db().managerModules().activeIdModule()
        if not idm:
            return

        if self.main_widget is None or self.main_widget.objectName() != idm:
            return

        windows_opened: List[str] = []
        if self._p_work_space is not None:
            for sub_window in self._p_work_space.subWindowList():
                child = sub_window.findChild(QtWidgets.QDialog)
                if child is not None:
                    windows_opened.append(child.idMDI())

        settings.SETTINGS.set_value("windowsOpened/%s" % idm, windows_opened)

        key = "Geometry/%s" % idm
        settings.SETTINGS.set_value("%s/Maximized" % key, self.main_widget.isMaximized())
        settings.SETTINGS.set_value("%s/X" % key, self.main_widget.x())
        settings.SETTINGS.set_value("%s/Y" % key, self.main_widget.y())
        settings.SETTINGS.set_value("%s/Width" % key, self.main_widget.width())
        settings.SETTINGS.set_value("%s/Height" % key, self.main_widget.height())

    def readState(self) -> None:
        """Read settings."""
        self._inicializing = False
        self._dict_main_widgets = {}

        if self.container_:
            rect_ = QtCore.QRect(self.container_.pos(), self.container_.size())
            self._multi_lang_enabled = settings.SETTINGS.value("MultiLang/Enabled", False)
            self._multi_lang_id = settings.SETTINGS.value(
                "MultiLang/LangId", QtCore.QLocale().name()[:2].upper()
            )

            if not settings.SETTINGS.value("Geometry/MainWindowMaximized", False):
                rect_.setX(settings.SETTINGS.value("Geometry/MainWindowX", rect_.x()))
                rect_.setY(settings.SETTINGS.value("Geometry/MainWindowY", rect_.y()))
                rect_.setWidth(settings.SETTINGS.value("Geometry/MainWindowWidth", rect_.width()))
                rect_.setHeight(
                    settings.SETTINGS.value("Geometry/MainWindowHeight", rect_.height())
                )

                frame_geo = self.frameGeometry()
                primary_screen = QtGui.QGuiApplication.primaryScreen()
                if primary_screen:
                    frame_geo.moveCenter(primary_screen.geometry().center())
                self.move(frame_geo.topLeft())

                desk = self.container_.frameGeometry()
                inter = desk.intersected(rect_)
                self.container_.resize(rect_.size())
                if inter.width() * inter.height() > (rect_.width() * rect_.height() / 20):
                    self.container_.move(rect_.topLeft())
            else:
                # FIXME: maximizar?
                self.container_.resize(self.container_.frameGeometry().size())

            active_id_module = self.db().managerModules().activeIdModule()

            windows_opened = settings.SETTINGS.value("windowsOpened/Main", [])

            for id_module in windows_opened:
                if id_module in self.db().managerModules().listAllIdModules():
                    widget = None
                    if id_module in self._dict_main_widgets.keys():
                        widget = self._dict_main_widgets[id_module]
                    if widget is None:
                        act = cast(
                            QtGui.QAction, self.container_.findChild(QtGui.QAction, id_module)
                        )
                        if not act or not act.isVisible():
                            continue

                        self.activateModule(id_module)

                        widget = self._dict_main_widgets[id_module]
                        widget.setObjectName(id_module)
                        if application.PROJECT.aq_app.acl_:
                            application.PROJECT.aq_app.acl_.process(widget)

                        self.setCaptionMainWidget("")
                        self.setMainWidget(widget)
                        application.PROJECT.aq_app.call("%s.init()" % id_module, [])
                        self.db().managerModules().setActiveIdModule(id_module)
                        self.setMainWidget(widget)
                        self.initMainWidget()

            for key, widget in self._dict_main_widgets.items():
                if widget.objectName() != active_id_module:
                    widget.installEventFilter(self)
                    widget.show()
                    widget.setFont(QtWidgets.QApplication.font())
                    if not isinstance(widget, QtWidgets.QMainWindow):
                        continue

                    view_back = widget.centralWidget()
                    if view_back is not None:
                        self._p_work_space = cast(
                            flworkspace.FLWorkSpace,
                            view_back.findChild(QtCore.QObject, widget.objectName()),
                        )

            if active_id_module:
                self.container_.show()
                self.container_.setFont(self.font())

            self.activateModule(active_id_module)

    def readStateModule(self) -> None:
        """Read settings for module."""

        mng_modules = self.db().managerModules()
        idm = mng_modules.activeIdModule()
        if not idm:
            return

        main_widget = self.main_widget
        if main_widget is None or main_widget.objectName() != idm:
            return

        self.read_state_widget(idm, main_widget)

        windows_opened = settings.SETTINGS.value("windowsOpened/%s" % idm, [])
        for action_name in windows_opened:
            action = cast(QtGui.QAction, main_widget.findChild(QtGui.QAction, action_name))
            if action and action.isVisible() and action_name in application.PROJECT.actions.keys():
                form = mng_modules.createForm(application.PROJECT.actions[action_name])
                self.read_state_widget(action_name, form)
                form.show()

    def read_state_widget(self, name: str, widget: QtWidgets.QWidget) -> None:
        """Read state from a widget."""

        key = "Geometry/%s" % name
        rect_ = QtCore.QRect(widget.pos(), widget.size())
        if not settings.SETTINGS.value("%s/Maximized" % key, False):
            rect_.setX(settings.SETTINGS.value("%s/X" % key, rect_.x()))
            rect_.setY(settings.SETTINGS.value("%s/Y" % key, rect_.y()))
            rect_.setWidth(settings.SETTINGS.value("%s/Width" % key, rect_.width()))
            rect_.setHeight(settings.SETTINGS.value("%s/Height" % key, rect_.height()))

            # desk = widget.frameGeometry()
            # inter = desk.intersected(rect_)
            widget.resize(rect_.size())
            widget.move(rect_.topLeft())

            # if (inter.width() * inter.height()) - 100 > (rect_.width() * rect_.height()):
            #    widget.move(rect_.topLeft())
            # else:
            #    widget.hide()
            #    widget.resize(desk.size())

    def __del__(self) -> None:
        """Cleanup."""
        self._destroying = True
        application.PROJECT.aq_app.stopTimerIdle()

        if self._dict_main_widgets:
            for key in list(self._dict_main_widgets.keys()):
                del self._dict_main_widgets[key]

            self._dict_main_widgets = {}

        self._ted_output = None

    def initView(self) -> None:
        """Initialize view."""

        if self.main_widget is None:
            return

        view_back = cast(QtWidgets.QMainWindow, self.main_widget).centralWidget()
        if not isinstance(view_back, QtWidgets.QMdiArea):
            view_back = QtWidgets.QMdiArea()
            view_back.setObjectName("mdi_area")
            view_back.setBackground(QtGui.QBrush(QtGui.QColor(255, 255, 255)))
            # view_back.logo = pixmap_fromMimeSource("pineboo-logo.png")
            # view_back.logo = aqs.AQS.pixmap_fromMimeSource("pineboo-logo.png")
            view_back.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
            view_back.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
            self._p_work_space = flworkspace.FLWorkSpace(
                view_back, self.db().managerModules().activeIdModule()
            )
            self._p_work_space.setAttribute(QtCore.Qt.WidgetAttribute.WA_NoSystemBackground)
            # p_work_space.setScrollBarsEnabled(True)
            # FIXME: setScrollBarsEnabled
            cast(QtWidgets.QMainWindow, self.main_widget).setCentralWidget(view_back)

    def setMainWidget(self, widget) -> None:
        """Set mainWidget."""
        if not self.container_:
            return

        application.PROJECT.aq_app.setMainWidget(widget)

        main_widget = (
            self.main_widget if isinstance(self.main_widget, QtWidgets.QMainWindow) else None
        )

        if main_widget is None:
            return

        tools_action = None
        status_action = None

        if self.toogle_bars_:
            tool_bar = cast(QtWidgets.QToolBar, main_widget.findChild(QtWidgets.QToolBar))
            for action in self.toogle_bars_.actions():
                if action.objectName() == "Herramientas":
                    tools_action = action
                elif action.objectName() == "Estado":
                    status_action = action

            if tool_bar and tools_action is not None:
                tools_action.setChecked(tool_bar.isVisible())

            if status_action is not None:
                status_action.setChecked(main_widget.statusBar().isVisible())  # type: ignore [union-attr]

    def showMainWidget(self, widget) -> None:
        """Show UI."""

        if not self.container_:
            if widget:
                widget.show()
            return

        focus_w = QtWidgets.QApplication.focusWidget()
        if widget is self.container_ or not widget:
            if self.container_.isMinimized():
                self.container_.showNormal()
            elif not self.container_.isVisible():
                self.container_.setFont(self.font())
                self.container_.show()

            if (
                focus_w
                and isinstance(focus_w, QtWidgets.QMainWindow)
                and focus_w != self.container_
            ):
                self.container_.setFocus()

            if not self.container_.isActiveWindow():
                self.container_.raise_()
                QtWidgets.QApplication.setActiveWindow(self.container_)

            if self.db() is not None:
                self.container_.setWindowTitle(self.db().mainConn().DBName())
            else:
                self.container_.setWindowTitle("Pineboo %s" % application.PROJECT.load_version())

            return

        if widget.isMinimized():
            widget.showNormal()
        elif not widget.isVisible():
            widget.show()
            widget.setFont(QtWidgets.QApplication.font())

        if focus_w and isinstance(focus_w, QtWidgets.QMainWindow) and focus_w != widget:
            widget.setFocus()
        if not widget.isActiveWindow():
            widget.raise_()
            QtWidgets.QApplication.setActiveWindow(widget)

        if widget:
            view_back = widget.centralWidget()
            if view_back:
                self._p_work_space = view_back.findChild(
                    flworkspace.FLWorkSpace, widget.objectName()
                )
                view_back.show()

        self.setCaptionMainWidget("")
        descript_area = (
            self.db()
            .managerModules()
            .idAreaToDescription(self.db().managerModules().activeIdArea())
        )
        widget.setWindowIcon(
            QtGui.QIcon(self.db().managerModules().iconModule(widget.objectName()))
        )
        self.tool_box_.setCurrentIndex(
            self.tool_box_.indexOf(self.tool_box_.findChild(QtWidgets.QToolBar, descript_area))
        )

    def initMainWidget(self) -> None:
        """Init mainwidget UI."""
        main_widget = cast(QtWidgets.QMainWindow, self.main_widget)
        if not main_widget or not self.container_:
            return

        if main_widget:
            action = main_widget.menuBar().addMenu(self.window_menu)  # type: ignore [union-attr]
            action.setText(self.tr("&Ventana"))  # type: ignore [union-attr]
            # main_widget.setCentralWidget(None)

        self.initView()
        self.initActions()
        self.initToolBar()
        self.initStatusBar()

        self.readStateModule()

    def setCaptionMainWidget(self, value: str) -> None:
        """Set application title."""
        if value:
            self.last_text_caption_ = value

        main_widget = self.main_widget

        if main_widget:
            descript_area = (
                self.db()
                .managerModules()
                .idAreaToDescription(self.db().managerModules().activeIdArea())
            )
            descript_module = self.db().managerModules().idModuleToDescription(self.objectName())

            if descript_area:
                main_widget.setWindowTitle(
                    "%s - %s - %s" % (self.last_text_caption_, descript_area, descript_module)
                )

    def initActions(self) -> None:
        """Initialize actions."""
        if self.main_widget is not None and self._p_work_space is not None:
            self.window_cascade_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                self._p_work_space.cascadeSubWindows
            )
            self.window_tile_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                self._p_work_space.tileSubWindows
            )
            self.window_close_action.triggered.connect(  # type: ignore [attr-defined] # noqa: F821
                self._p_work_space.closeActiveSubWindow
            )

    def initStatusBar(self) -> None:
        """Initialize statusbar."""
        main_widget = self.main_widget

        if not main_widget:
            return

        application.PROJECT.aq_app.statusHelpMsg(self.tr("Listo."))

        if main_widget is not None:
            status_bar = cast(QtWidgets.QMainWindow, main_widget).statusBar()
            if status_bar:
                status_bar.setSizeGripEnabled(False)

                conexion = QtWidgets.QLabel(status_bar)
                conexion.setText("%s@%s" % (self.db().user(), self.db().DBName()))
                status_bar.addWidget(conexion)

    def toggleToolBar(self, toggle: bool) -> None:
        """Show or hide toolbar."""
        main_widget = cast(QtWidgets.QToolBar, self.main_widget)

        if not main_widget:
            return

        tool_bar = cast(QtWidgets.QToolBar, main_widget.findChild(QtWidgets.QToolBar))

        if not tool_bar:
            return

        if toggle:
            tool_bar.show()
        else:
            tool_bar.hide()

    def toggleStatusBar(self, toggle: bool) -> None:
        """Toggle status bar."""
        main_widget = cast(QtWidgets.QMainWindow, self.main_widget)
        if not main_widget:
            return
        if toggle:
            main_widget.statusBar().show()  # type: ignore [union-attr]
        else:
            main_widget.statusBar().hide()  # type: ignore [union-attr]

    def generalExit(self, ask_exit=True) -> bool:
        """Perform before close checks."""
        do_exit = True
        if ask_exit:
            do_exit = application.PROJECT.aq_app.queryExit()
        if do_exit:
            self._destroying = True
            # if application.PROJECT.aq_app.consoleShown():
            #    if application.PROJECT.aq_app._ted_output is not None:
            #        application.PROJECT.aq_app._ted_output.close()

            if not application.PROJECT.aq_app.form_alone_:
                self.writeState()
                self.writeStateModule()

            if self.db().driverName():
                self.db().managerModules().finish()
                self.db().manager().finish()
                # QtCore.QTimer.singleShot(0, application.PROJECT.aq_app.quit)

            for main_widget in self._dict_main_widgets.values():
                main_widget.close()

            return True
        else:
            return False



"""
Project Module.
"""

from pineboolib import logging

from pineboolib.core.utils import utils_base, struct
from pineboolib.core import exceptions, settings, message_manager, decorators

from pineboolib.application.database import pnconnectionmanager
from pineboolib.application.database import utils as db_utils
from pineboolib.application.parsers.parser_mtd import pnmtdparser, pnormmodelsfactory
from pineboolib.application.parsers import parser_qsa
from pineboolib.application.utils import path, xpm, flfiles_dir
from pineboolib.application import qsadictmodules

from pineboolib.application import module, file as file_module
from pineboolib.application import connections

import os
from typing import List, Optional, Any, Dict, Callable, Union, TYPE_CHECKING

from pineboolib import application

if TYPE_CHECKING:
    from pineboolib.interfaces import (
        dgi_schema,
        imainwindow,
        iconnection,
    )  # noqa: F401 # pragma: no cover
    from pineboolib.application.database import pnconnection  # pragma: no cover
    from pineboolib.application import xmlaction, pnapplication  # noqa: F401 # pragma: no cover
    from PyQt6 import QtWidgets  # type: ignore[import] # pragma: no cover
    from pineboolib.qsa import formdbwidget


LOGGER = logging.get_logger(__name__)


class Project(object):
    """
    Singleton for the whole application.

    Can be accessed with pineboolib.project from anywhere.
    """

    _conn_manager: Optional["pnconnectionmanager.PNConnectionManager"]

    _app: "QtWidgets.QApplication"
    _aq_app: Optional["pnapplication.PNApplication"] = None

    main_window: Optional["imainwindow.IMainWindow"] = None
    dgi: Optional["dgi_schema.dgi_schema"] = None
    delete_cache: bool = False
    delete_base_cache: bool
    path = None
    _splash = None
    sql_drivers_manager = None
    timer_ = None
    no_python_cache: bool
    _msg_mng = None
    alternative_folder: Optional[str]
    _session_func_: Optional["Callable"]

    areas: Dict[str, "struct.AreaStruct"]
    files: Dict[str, "file_module.File"]
    actions: Dict[str, "xmlaction.XMLAction"]

    modules: Dict[str, "module.Module"]
    pending_conversion_list: List[str]
    USE_FLFILES_FOLDER: str = ""
    db_admin_mode: bool = False

    def __init__(self) -> None:
        """Initialize."""
        # self._conn = None
        self.dgi = None
        self.tree = None
        self.root = None
        self.alternative_folder = None
        self.apppath = ""
        self.tmpdir = settings.CONFIG.value("ebcomportamiento/temp_dir", "")
        self.parser = None

        self.no_python_cache = settings.CONFIG.value("ebcomportamiento/noPythonCache", False)
        self.apppath = utils_base.filedir("..")
        self.delete_base_cache = settings.CONFIG.value("ebcomportamiento/keep_general_cache", False)
        self.delete_cache = settings.CONFIG.value("ebcomportamiento/deleteCache", False)
        self.actions = {}
        self.files = {}
        self.areas = {}
        self.modules = {}
        self.db_admin_mode = settings.CONFIG.value("application/dbadmin_enabled", False)

        import pathlib

        if not self.tmpdir:
            self.tmpdir = utils_base.filedir("%s/Pineboo/tempdata" % pathlib.Path.home())
            settings.CONFIG.set_value("ebcomportamiento/temp_dir", self.tmpdir)

        if not os.path.exists(self.tmpdir):
            try:
                pathlib.Path(self.tmpdir).mkdir(parents=True, exist_ok=True)
            except Exception as error:
                LOGGER.error("Error creating %s folder : %s", self.tmpdir, str(error))
                return

        if not os.access(self.tmpdir, os.W_OK):
            LOGGER.error("%s folder is not writable!. Please change permissions!", self.tmpdir)
            return

        self._conn_manager = None
        self._session_func_ = None
        LOGGER.debug("Initializing connection manager for the application.PROJECT %s", self)
        self.pending_conversion_list = []

    @property
    def app(self) -> "QtWidgets.QApplication":
        """Retrieve current Qt Application or throw error."""
        if self._app is None:
            raise Exception("No application set")
        return self._app

    def set_app(self, app: "QtWidgets.QApplication"):
        """Set Qt Application."""
        self._app = app

    @property
    def aq_app(self) -> "pnapplication.PNApplication":
        """Retrieve current Qt Application or throw error."""
        if self._aq_app is None:
            from pineboolib.application import pnapplication

            self._aq_app = pnapplication.PNApplication()
        return self._aq_app

    def set_aq_app(self, aq_app: "pnapplication.PNApplication") -> None:
        """Set Qt Application."""
        self._aq_app = aq_app

    @property
    def conn_manager(self) -> "pnconnectionmanager.PNConnectionManager":
        """Retrieve current connection or throw."""
        if self._conn_manager is None:
            self._conn_manager = pnconnectionmanager.PNConnectionManager()

        return self._conn_manager

    @property
    def DGI(self) -> "dgi_schema.dgi_schema":
        """Retrieve current DGI or throw."""
        if self.dgi is None:
            raise Exception("Project is not initialized")
        return self.dgi

    def init_conn(self, connection: "pnconnection.PNConnection") -> bool:
        """Initialize project with a connection."""
        # if self._conn is not None:
        #    del self._conn
        #    self._conn = None

        return self.conn_manager.setMainConn(connection)

    def init_dgi(self, dgi: "dgi_schema.dgi_schema") -> None:
        """Load and associate the defined DGI onto this project."""
        # FIXME: Actually, DGI should be loaded here, or kind of.

        self.dgi = dgi

        self._msg_mng = message_manager.Manager(dgi)

        self.dgi.extraProjectInit()

    def load_modules(self) -> None:
        """Load all modules."""
        for module_name, mod_obj in self.modules.items():
            mod_obj.load()

    def load_orm(self) -> None:
        """Load Orm objects."""

        for file_item in [item for item in self.files.values() if item.filename.endswith(".mtd")]:
            file_name_model = "%s_model.py" % file_item.filename[:-4]
            if file_name_model not in self.files.keys():
                path_file = pnmtdparser.mtd_parse(file_item.filename, file_item.path())
                if path_file:
                    self.files[file_name_model] = file_module.File(
                        file_item.module,
                        "%s_model.py" % file_item.path(),
                        basedir=file_item.basedir,
                        sha=file_item.sha,
                        db_name=self.conn_manager.mainConn().DBName(),
                    )

                    self.files[file_name_model].filekey = "%s_model.py" % file_item.filekey

            else:
                LOGGER.debug(
                    "%s already exists (%s).", file_name_model, self.files[file_name_model].path()
                )

        self.message_manager().send("splash", "showMessage", ["Cargando objetos ..."])
        LOGGER.info("Loading ORMS ...")
        pnormmodelsfactory.load_models()

    def load_classes(self) -> None:
        """Load class files into qsa tree."""

        for key in list(self.files.keys()):
            if not key.endswith(".py") or key.startswith("test_"):
                continue

            db_utils.process_file_class(self.files[key])

    def run(self) -> bool:
        """Run project. Connects to DB and loads data."""

        LOGGER.info("RUN: Loading project data.")

        self.pending_conversion_list.clear()
        self.actions.clear()
        self.files.clear()
        self.areas.clear()
        self.modules.clear()

        if self.dgi is None:
            raise Exception("DGI not loaded")

        delete_cache = self.delete_cache
        cache_ver = parser_qsa.PARSER_QSA_VERSION

        cache_folder = path._dir("cache")
        db_cache_folder = os.path.join(cache_folder, self.conn_manager.mainConn().DBName())
        cache_version_file_path = os.path.join(db_cache_folder, "cache_version.txt")

        if not os.path.exists(cache_folder):
            path_build: List[str] = []
            try:
                LOGGER.info("RUN: Checking if cache folder exists (%s)", cache_folder)
                for folder in os.path.split(cache_folder):
                    path_build.append(folder)
                    path_med = os.path.join(*path_build)
                    if not os.path.exists(path_med):
                        os.mkdir(path_med)
            except Exception as error:
                raise Exception(
                    "Error building cache folder (%s) : %s" % (path_build, error)
                ) from error

        if os.path.exists(db_cache_folder):
            if not os.path.exists(cache_version_file_path):
                delete_cache = True
            else:
                cache_ver = ""
                try:
                    file_ver = open(cache_version_file_path, "r", encoding="UTF8")
                    cache_ver = file_ver.read()
                    file_ver.close()
                except Exception:
                    pass
                if cache_ver != parser_qsa.PARSER_QSA_VERSION:
                    delete_cache = True

            if delete_cache:
                if cache_ver != parser_qsa.PARSER_QSA_VERSION:
                    LOGGER.warning(
                        "QSA parser version has changed from %s to %s!. Deleting cache.",
                        cache_ver,
                        parser_qsa.PARSER_QSA_VERSION,
                    )
                else:
                    LOGGER.warning("Deleting cache.")

        if delete_cache and os.path.exists(db_cache_folder):
            self.message_manager().send("splash", "showMessage", ["Borrando caché ..."])
            LOGGER.info("DEVELOP: delete_cache Activado\nBorrando %s", db_cache_folder)

            for root, dirs, files in os.walk(db_cache_folder, topdown=False):
                for name in files:
                    if os.path.exists(os.path.join(root, name)):
                        os.remove(os.path.join(root, name))
                for name in dirs:
                    if name != "sqlite_database":
                        if os.path.exists(os.path.join(root, name)):
                            os.rmdir(os.path.join(root, name))

        elif self.delete_base_cache:
            for file_name in os.listdir(self.tmpdir):
                if file_name.find(".") > -1 and not file_name.endswith("sqlite3"):
                    file_path = os.path.join(self.tmpdir, file_name)
                    try:
                        os.remove(file_path)
                    except Exception:
                        LOGGER.warning("No se ha podido borrar %s al limpiar la cache", file_path)
                        pass

        if not os.path.exists(cache_folder):
            LOGGER.info("RUN: Creating %s folder.", cache_folder)
            os.makedirs(cache_folder)

        if not os.path.exists(db_cache_folder):
            LOGGER.info("RUN: Creating %s folder.", db_cache_folder)
            os.makedirs(db_cache_folder)

        file_ver = open(cache_version_file_path, "w", encoding="UTF8")
        file_ver.write(parser_qsa.PARSER_QSA_VERSION)
        file_ver.close()
        del file_ver

        return self.load_system_module() and self.load_database_modules()

    def call(
        self,
        function: str,
        args: List[Any],
        object_context: Union["formdbwidget.FormDBWidget", "object", None] = None,
        show_exceptions: bool = True,
        default_value: Any = True,
    ) -> Optional[Any]:
        """
        Call to a QS project function.

        @param function. Nombre de la función a llamar.
        @param args. Array con los argumentos.
        @param object_context. Contexto en el que se ejecuta la función.
        @param show_exceptions. Boolean que especifica si se muestra los errores.
        @return Boolean con el resultado.
        """

        LOGGER.trace(
            "JS.CALL: fn:%s args:%s ctx:%s", function, args, object_context, stack_info=True
        )

        if not application.ENABLE_CALL_EXCEPTIONS:
            show_exceptions = False

        # Tipicamente flfactalma.iface.beforeCommit_articulos()
        if function[-2:] == "()":
            function = function[:-2]

        array_fun = function.split(".")
        module_name = array_fun[0]
        function_name = array_fun[-1]

        if object_context is None:
            object_context = qsadictmodules.QSADictModules.from_project(module_name)

            if hasattr(
                object_context, "iface"
            ) and hasattr(  # comprueba si la función es realmente de iface.
                object_context.iface, function_name  # type: ignore [union-attr]
            ):
                object_context = object_context.iface  # type: ignore [union-attr]

        function_object = getattr(object_context, function_name, None)
        if function_object is not None:
            try:
                # Controlar numero de argumentos
                args_num = connections.get_expected_args_num(function_object)
                while args_num > len(args):
                    args.append(None)

                if args and args_num:
                    args = args[0:args_num]

                return function_object(*args)
            except Exception as error:
                if show_exceptions:
                    LOGGER.exception(
                        "JSCALL: Error executing function %s ERROR: %s" % (function_name, error),
                        stack_info=True,
                    )
        else:
            if show_exceptions:
                LOGGER.error("No existe la función %s en %s", function_name, module_name)
            return default_value
            # FIXME: debería ser false, pero igual se usa por el motor para detectar propiedades

        return None

    def parse_script(self, scriptname: str, txt_: str = "") -> bool:
        """
        Convert QS script into Python and stores it in the same folder.

        @param scriptname, Nombre del script a convertir
        """
        from pineboolib.application.parsers.parser_qsa import postparse

        # Intentar convertirlo a Python primero con flscriptparser2
        if not os.path.isfile(scriptname):
            raise IOError
        python_script_path = (scriptname + ".xml.py").replace(".qs.xml.py", ".qs.py")
        if not os.path.isfile(python_script_path) or self.no_python_cache:
            file_name_l = scriptname.split(os.sep)  # FIXME: is a bad idea to split by os.sep
            file_name = file_name_l[len(file_name_l) - 2]

            msg = "Convirtiendo a Python . . . %s.qs %s" % (file_name, txt_)
            LOGGER.info(msg)

            try:
                postparse.pythonify([scriptname], ["--strict"])
            except Exception as error:
                LOGGER.exception(
                    "El fichero %s no se ha podido convertir: %s", scriptname, str(error)
                )
                return False

        return True

    def parse_script_list(self, path_list: List[str]) -> bool:
        """Convert QS scripts list into Python and stores it in the same folders."""

        from pineboolib.application.parsers.parser_qsa import pyconvert

        if not path_list:
            return True

        for file_path in path_list:
            if not os.path.isfile(file_path):
                raise IOError

        itemlist = []
        size_list = len(path_list)
        for num, orig_file_name in enumerate(path_list):
            dest_file_name = "%s.py" % orig_file_name[:-3]
            if dest_file_name in self.pending_conversion_list:
                LOGGER.warning("The file %s is already being converted. Waiting", dest_file_name)
                while dest_file_name in self.pending_conversion_list:
                    # Esperamos a que el fichero se convierta.
                    self.app.processEvents()  # type: ignore[misc] # noqa: F821
            else:
                self.pending_conversion_list.append(dest_file_name)
                itemlist.append(
                    pyconvert.PythonifyItem(
                        src=orig_file_name, dst=dest_file_name, number=num, len=size_list, known={}
                    )
                )

        threads_num = pyconvert.CPU_COUNT
        if len(itemlist) < threads_num:
            threads_num = len(itemlist)

        pycode_list: List[bool] = []

        if parser_qsa.USE_THREADS:
            import multiprocessing

            with multiprocessing.Pool(threads_num) as thread:
                # TODO: Add proper signatures to Python files to avoid reparsing
                pycode_list = thread.map(pyconvert.pythonify_item, itemlist, chunksize=2)
        else:
            for item in itemlist:
                pycode_list.append(pyconvert.pythonify_item(item))

        for item in itemlist:
            self.pending_conversion_list.remove(item.dst_path)

        if not all(pycode_list):
            LOGGER.warning("Conversion failed for some files")
            return False
        # LOGGER.warning("Parseados %s", path_list)
        return True

    @decorators.deprecated
    def get_temp_dir(self) -> str:
        """
        Return temporary folder defined for pineboo.

        @return ruta a la carpeta temporal
        """
        # FIXME: anti-pattern in Python. Getters for plain variables are wrong.
        raise exceptions.CodeDoesNotBelongHereException("Use project.tmpdir instead, please.")
        # return self.tmpdir

    def load_version(self) -> str:
        """Initialize current version numbers."""
        from pineboolib.application import PINEBOO_VER

        return "DBAdmin v%s" % PINEBOO_VER if self.db_admin_mode else "Quick v%s" % PINEBOO_VER

    def message_manager(self):
        """Return message manager for splash and progress."""
        return self._msg_mng

    def set_session_function(self, fun_: Callable) -> None:
        """Set session funcion."""

        self._session_func_ = fun_

    def session_id(self) -> str:
        """Return id if use pineboo like framework."""

        return str(self._session_func_()) if self._session_func_ is not None else "auto"

    def load_system_module(self) -> bool:
        """Load system module."""

        base_dir = utils_base.get_base_dir()
        is_library = utils_base.is_library()

        file_object = open(
            utils_base.filedir(base_dir, "system_module", "sys.xpm"), "r", encoding="UTF-8"
        )
        icono = file_object.read()
        file_object.close()

        del file_object

        self.modules["sys"] = module.Module("sys", "sys", "Administración", icono, "1.0")
        for root, dirs, files in os.walk(utils_base.filedir(base_dir, "system_module")):
            for nombre in files:
                if "tests" in root:
                    continue

                if is_library and nombre.endswith("ui"):
                    continue

                if nombre.endswith("__.py") or nombre.endswith(".src"):
                    continue

                if root.find("modulos") == -1:
                    fileobj = file_module.File(
                        "sys", nombre, basedir=root, db_name=self.conn_manager.mainConn().DBName()
                    )
                    self.files[nombre] = fileobj
                    self.modules["sys"].add_project_file(fileobj)
                    del fileobj

        pnormmodelsfactory.load_models()
        # Se verifica que existen estas tablas
        for table in (
            "flareas",
            "flmodules",
            "flfiles",
            "flgroups",
            "fllarge",
            "flserial",
            "flusers",
            "flvar",
            "flmetadata",
            "flsettings",
            "flupdates",
            "flseqs",
        ):
            self.conn_manager.manager().createSystemTable(table)

        return True

    def load_database_modules(self) -> bool:
        """Load database modules."""

        conn = self.conn_manager.dbAux()
        db_name = conn.DBName()
        is_library = utils_base.is_library()
        result_areas: Any = []
        static_flfiles = None

        if self.USE_FLFILES_FOLDER:
            LOGGER.warning("FLFILES_FOLDER: Using %s like flfiles", self.USE_FLFILES_FOLDER)
            static_flfiles = flfiles_dir.FlFiles(self.USE_FLFILES_FOLDER)
            result_areas = static_flfiles.areas()
        else:
            result_areas = conn.execute_query(
                """SELECT idarea, descripcion FROM flareas WHERE 1 = 1"""
            )

        for idarea, descripcion in list(result_areas):
            if idarea == "sys":
                continue
            self.areas[idarea] = struct.AreaStruct(idarea=idarea, descripcion=descripcion)

        self.areas["sys"] = struct.AreaStruct(idarea="sys", descripcion="Area de Sistema")

        result_modules: Any = []
        # Obtener módulos activos
        if static_flfiles:
            result_modules = static_flfiles.modules()
        else:
            result_modules = conn.execute_query(
                """SELECT idarea, idmodulo, descripcion, icono, version FROM flmodules WHERE bloqueo = %s """
                % conn.driver().formatValue("bool", "True", False)
            )

        for idarea, idmodulo, descripcion, icono, version in list(result_modules):
            if idmodulo not in self.modules:
                icon_cached = xpm.cache_xpm(icono)
                self.modules[idmodulo] = module.Module(
                    idarea, idmodulo, descripcion, icon_cached, version
                )

        result_files: Any = []
        if static_flfiles:
            result_files = static_flfiles.files()
        else:
            result_files = conn.execute_query(
                """SELECT idmodulo, nombre, sha, bloqueo FROM flfiles WHERE NOT sha = '' ORDER BY idmodulo, nombre """
            )

        log_file = open(path._dir("project.txt"), "w", encoding="UTF-8")

        list_files: List[str] = []
        LOGGER.info("RUN: Populating cache.")
        for idmodulo, nombre, sha, contenido_or_bloqueo in list(result_files):
            if idmodulo not in self.modules.keys():  # Si el módulo no existe.
                continue

            elif is_library and nombre.endswith("ui"):  # Si es un UI en modo librería.
                continue

            elif nombre in self.files.keys():  # Si se sobreescribe un fichero ya existente.
                if self.files[nombre].module == "sys":
                    continue
                else:
                    LOGGER.warning("run: file %s already loaded, overwritting..." % nombre)

            fileobj = file_module.File(idmodulo, nombre, sha, db_name=db_name)
            self.files[nombre] = fileobj

            self.modules[idmodulo].add_project_file(fileobj)

            log_file.write(fileobj.filekey + "\n")

            fileobjdir = os.path.dirname(path._dir("cache", fileobj.filekey))
            file_name = path._dir("cache", fileobj.filekey)
            if not os.path.isfile(file_name) or not os.path.getsize(
                file_name
            ):  # Borra si no existe el fichero o está vacio.
                if os.path.exists(fileobjdir):
                    utils_base.empty_dir(fileobjdir)
                else:
                    os.makedirs(fileobjdir, exist_ok=True)

                contenido_content: Optional[str] = None

                if not static_flfiles:
                    qry = conn.execute_query(
                        """SELECT contenido FROM flfiles WHERE sha = %s AND nombre = %s """
                        % (
                            conn.driver().formatValue("string", sha, False),
                            conn.driver().formatValue("string", nombre, False),
                        )
                    )

                    result_content: Any = None
                    if qry:
                        result_content = qry.first()

                    if result_content is not None:
                        contenido_content = result_content[
                            0
                        ]  # Recogemos verdadero contenido_content. cuando usamos flfiles. más rpapido conexiones lentas.
                else:
                    contenido_content = contenido_or_bloqueo

                if contenido_content is not None:
                    encode_ = "UTF-8" if str(nombre).endswith((".ts", ".py")) else "ISO-8859-15"
                    self.message_manager().send(
                        "splash", "showMessage", ["Volcando a caché %s..." % nombre]
                    )

                    new_cache_file = open(file_name, "wb")
                    new_cache_file.write(contenido_content.encode(encode_, "replace"))
                    new_cache_file.close()
            else:
                if file_name.endswith(".py"):
                    static_flag = "%s/static.xml" % fileobjdir
                    if os.path.exists(static_flag):
                        os.remove(static_flag)

            if application.PARSE_PROJECT_ON_INIT:
                if nombre.endswith(".qs"):
                    if self.no_python_cache or not os.path.exists(
                        "%spy" % file_name[:-2]
                    ):  # si es forzado o no existe el .py
                        list_files.append(file_name)

        log_file.close()
        LOGGER.info("RUN: End populating cache.")
        if self.USE_FLFILES_FOLDER and application.UPDATE_FLFILES_FROM_FLFOLDER:
            self.update_flfiles(result_files, conn)

        self.conn_manager.removeConn("dbaux")
        del log_file

        self.message_manager().send(
            "splash",
            "showMessage",
            ["Convirtiendo a Python %s ..." % ("(forzado)" if self.no_python_cache else "")],
        )
        if list_files:
            LOGGER.info("RUN: Parsing QSA files. (%s): %s", len(list_files), list_files)
            if not self.parse_script_list(list_files):
                LOGGER.warning("Failed QSA conversion !.See debug for mode information.")
                return False

        return True

    def update_flfiles(self, files_list: List[List[str]], conn: "iconnection.IConnection") -> None:
        """Actualiza la tabla flfiles con el contenido de flfiles_folder."""

        model_flsettins = qsadictmodules.QSADictModules.orm_("flsettings").get("sha_flfiles")
        sha_flfiles = model_flsettins.valor if model_flsettins else None
        sha_flfolder = ""
        for id_module, file_name, string_sha, data in files_list:
            sha_flfolder = utils_base.sha1("%s%s" % (sha_flfolder, string_sha))

        if not sha_flfolder or sha_flfolder == sha_flfiles:
            LOGGER.warning("SHA_FLFILES:%s" % (sha_flfiles))
            return
        LOGGER.warning("SHA_FLFILES CHANGED: OLD:%s, NEW:%s" % (sha_flfiles, sha_flfolder))
        LOGGER.warning("Updating flfiles table from flfiles folder.")
        # 1 vaciar flfiles
        LOGGER.warning("(1/5) Deleting older data from tables ...")
        conn.execute_query("DELETE FROM flareas")
        conn.execute_query("DELETE FROM flmodules")
        conn.execute_query("DELETE FROM flfiles")
        # 2 insertar flareas
        LOGGER.warning("(2/5) Updating flareas ...")
        for data in self.areas.values():
            model_areas = qsadictmodules.QSADictModules.orm_("flareas")()
            model_areas.idarea = data.idarea
            model_areas.descripcion = data.descripcion
            model_areas.bloqueo = data.idarea != "sys"
            if not model_areas.save():
                LOGGER.error("Error saving area %s", data.idarea)
                continue
        LOGGER.warning("(3/5) Updating flmodules ...")
        for data in self.modules.values():
            model_modules = qsadictmodules.QSADictModules.orm_("flmodules")()
            model_modules.idarea = data.areaid
            model_modules.idmodulo = data.name
            model_modules.descripcion = data.description
            model_modules.icono = data.icon
            model_modules.version = data.version
            model_modules.bloqueo = data.name != "sys"
            if not model_modules.save():
                LOGGER.error("Error saving module %s", data.name)
                continue

        LOGGER.warning("(4/5) Updating flfiles ...")
        for id_module, file_name, string_sha, data in files_list:
            if id_module == "sys":
                continue

            if not file_name.endswith(
                (
                    ".ar",
                    ".jrxml",
                    ".kut",
                    ".mod",
                    ".mtd",
                    ".py",
                    ".qs",
                    ".qry",
                    ".ts",
                    ".ui",
                    ".xml",
                    ".xpm",
                )
            ):
                LOGGER.debug("\t* Skipping %s", file_name)
                continue

            model_files = qsadictmodules.QSADictModules.orm_("flfiles")()
            model_files.idmodulo = id_module
            model_files.nombre = file_name
            model_files.sha = string_sha
            model_files.contenido = data
            result = False
            error = ""
            try:
                result = model_files.save()
            except Exception as err:
                error = str(error)

            if not result:
                LOGGER.error("Error saving file %s. %s", file_name, error)

        # 6 Limpiar/actualizar flmetadata
        LOGGER.warning("(5/5) Updating flmetadata ...")
        self.conn_manager.manager().cleanupMetaData()
        # 7 actualizar sha
        # flutil.FLUtil.writeDBSettingEntry("sha_flfiles", sha_flfolder)
        model_settings = (
            qsadictmodules.QSADictModules.orm_("flsettings").get("sha_flfiles")
            or qsadictmodules.QSADictModules.orm_("flsettings")()
        )
        model_settings.flkey = "sha_flfiles"
        model_settings.valor = sha_flfolder
        if not model_settings.save():
            LOGGER.error("Error saving sha_flfiles")
            return

        LOGGER.warning("Update completed. New sha_flfiles:%s" % (sha_flfolder))



# -*- coding: utf-8 -*-
"""About_pineboo module."""

from PyQt6 import QtWidgets, QtCore  # type: ignore[import]

from pineboolib.core.utils import check_dependencies

import platform

from typing import Any


class AboutPineboo(QtWidgets.QDialog):
    """AboutPineboo class."""

    def __init__(self) -> None:
        """Inicialize."""

        super().__init__()
        self.load()

    def load(self) -> None:
        """Load widget and show."""

        from pineboolib import application
        from pineboolib.core.utils.utils_base import filedir

        dlg_ = filedir("plugins/dgi/dgi_qt/dgi_objects/dlg_about/about_pineboo.ui")
        version_ = application.PROJECT.load_version()
        self.ui_: Any = application.PROJECT.conn_manager.managerModules().createUI(dlg_, None, self)
        if self.ui_ is None:
            raise Exception("Error creating UI About Dialog")

        self.ui_.lbl_version.setText("Pineboo %s" % str(version_))
        self.ui_.btn_close.clicked.connect(self.ui_.close)
        self.ui_.btn_clipboard.clicked.connect(self.to_clipboard)
        self.ui_.show()

        self.ui_.lbl_librerias.setText(self.load_components())

    def load_components(self) -> str:
        """Resume libraries loaded."""

        components = "Versiones de componentes:\n\n"
        components += "S.O.: %s %s %s\n" % (
            platform.system(),
            platform.release(),
            platform.version(),
        )

        if "PyQt6.QtCore" not in check_dependencies.DEPENDENCIES_CHECKED.keys():
            components += "PyQt6.QtCore: %s\n" % QtCore.QT_VERSION_STR

        for k in check_dependencies.DEPENDENCIES_CHECKED.keys():
            components += "%s: %s\n" % (k, check_dependencies.DEPENDENCIES_CHECKED[k])

        return components

    def to_clipboard(self) -> None:
        """Copy resume libraries loaded into clipboard."""

        clip_board = QtWidgets.QApplication.clipboard()
        if clip_board:
            clip_board.clear()
            clip_board.setText(self.load_components())



"""
Manage buffers used by PNSqlCursor.

*What is a buffer?*

Buffers are the data records pointed to by a PNSqlCursor.
"""

from pineboolib.application import types
from pineboolib.application.database import utils as utils_database
from pineboolib import logging
from pineboolib.core.utils import utils_base
from pineboolib.application.utils import xpm

import datetime
import sqlalchemy  # type: ignore [import]

from typing import List, Union, Optional, Callable, Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import isqlcursor  # pragma: no cover
    from pineboolib.application.database import pncursortablemodel  # pragma: no cover
    import decimal  # noqa : F821 # pragma: no cover

LOGGER = logging.get_logger(__name__)

ACCEPTABLE_VALUES = (
    int,
    float,
    str,
    "datetime.time",
    "datetime.date",
    bool,
    types.Date,
    bytearray,
    "decimal.Decimal",
    "datetime.timedelta",
)
TVALUES = Union[
    int,
    float,
    str,
    "datetime.time",
    "datetime.date",
    bool,
    types.Date,
    bytearray,
    "datetime.timedelta",
    None,
    Dict[Any, Any],
]


class PNBuffer(object):
    """
    Cursor buffer.

    When a query is done, after first(), a PNBuffer is created which holds
    the fields of the record.
    """

    _orm_obj: Optional[Callable]
    _generated_fields: List[str]
    _cache_buffer: Dict[str, TVALUES]
    _cursor: "isqlcursor.ISqlCursor"
    _init_orm: bool

    def __init__(self, cursor: "isqlcursor.ISqlCursor") -> None:
        """Create a Buffer from the specified PNSqlCursor."""
        super().__init__()
        if not cursor:
            raise Exception("Missing cursor")
        self._cursor = cursor
        self._orm_obj = None
        self._generated_fields = []
        self._cache_buffer = {}
        self._init_orm = True

    def prime_insert(self, row: Optional[int] = None) -> None:
        """
        Set the initial values of the buffer fields.

        @param row = cursor line.
        """
        self.clear()
        self._orm_obj = self._cursor._cursor_model(
            session=self._cursor.db().session(), no_init=not self._init_orm
        )

    def prime_update(self) -> None:
        """Set the initial copy of the cursor values into the buffer."""

        self.clear()
        self._orm_obj = self.model().get_obj_from_row(self._cursor.currentRegister())

    def setNull(self, name) -> None:
        """
        Empty the value of the specified field.

        @param name = field name.
        """
        setattr(self._orm_obj, name, None)

    def value(self, field_name: str, return_none: bool = False) -> "TVALUES":
        """
        Return the value of a field.

        @param field_name field identification.
        @return Any = field value.
        """

        if field_name in self._cache_buffer.keys():
            value = self._cache_buffer[field_name]
        else:
            if self._orm_obj and sqlalchemy.inspect(self._orm_obj).expired:
                self._orm_obj = self.model().get_obj_from_row(self._cursor.currentRegister())

            value = getattr(self._orm_obj, field_name, None)

        metadata = self._cursor.metadata().field(field_name)

        if metadata is not None:
            type_ = metadata.type()

            if value is not None:
                if type_ == "pixmap":
                    v_large = (
                        xpm.cache_xpm(str(value))
                        if self._cursor.private_cursor._is_system_table
                        else self._cursor.db().connManager().manager().fetchLargeValue(str(value))
                    )

                    value = v_large if v_large else value
                else:
                    value = utils_database.resolve_qsa_value(type_, value)
            elif return_none is False:
                value = utils_database.resolve_empty_qsa_value(type_)

        return value

    def set_value(self, field_name: str, value: "TVALUES") -> bool:
        """Set values to cache_buffer."""

        if field_name not in self._cursor.metadata().fieldNames():
            return False

        self._cache_buffer[field_name] = value
        return True

    def apply_buffer(self) -> bool:
        """Aply buffer to object (commitBuffer)."""
        ret_ = True

        for field_name in self._cache_buffer.keys():
            value: Any = self._cache_buffer[field_name]
            ret_ = self.set_value_to_objet(field_name, value)
            if not ret_:
                break

        return ret_

    def set_value_to_objet(self, field_name: str, value: "TVALUES") -> bool:
        """
        Set the value of a field.

        @param name = Field name.
        @param value = new value.
        @param mark_. If True verifies that it has changed from the value assigned in primeUpdate and mark it as modified (Default to True).
        """
        if value is not None:
            metadata = self._cursor.metadata().field(field_name)
            if metadata is not None:
                type_ = metadata.type()

                if type_ == "double":
                    if isinstance(value, str) and value == "":
                        value = None
                    else:
                        value = float(value)  # type: ignore [arg-type]
                elif type_ in ("int", "uint", "serial"):
                    if isinstance(value, str) and value == "":
                        value = None
                    else:
                        value = int(value)  # type: ignore [arg-type]
                elif type_ in ("string", "pixmap", "stringlist", "counter"):
                    value = str(value)
                elif type_ in ("boolean", "unlock"):
                    value = utils_base.text2bool(str(value))

                if value in ["", "NULL"]:
                    if isinstance(value, str) and value == "NULL":
                        value = None
                else:
                    if type_ == "date":
                        value = datetime.datetime.strptime(str(value)[:10], "%Y-%m-%d")
                    elif type_ == "timestamp":
                        value = datetime.datetime.strptime(str(value)[0:19], "%Y-%m-%d %H:%M:%S")
                    elif type_ == "time":
                        value = str(value)
                        if value.find("T") > -1:
                            value = value[value.find("T") + 1 :]

                        value = datetime.datetime.strptime(str(value)[:8], "%H:%M:%S").time()
                    elif type_ in ["bool", "unlock"]:
                        value = True if value in [True, 1, "1", "true"] else False

        try:
            setattr(self._orm_obj, field_name, value)
        except Exception as error:
            LOGGER.error("setValue: %s", str(error))
            return False
        return True

    def current_object(self) -> "Callable":
        """Return current db object."""

        if not self._orm_obj:
            raise Exception("buffer orm object doesn't exists!!")

        return self._orm_obj

    def model(self) -> "pncursortablemodel.PNCursorTableModel":
        """Return cursor table model."""

        return self._cursor.model()

    def clear(self):
        """Clear buffer object."""

        del self._orm_obj
        self._orm_obj = None
        del self._cache_buffer
        self._cache_buffer = {}

    def is_null(self, field_name: str) -> bool:
        """Return if a field is null."""

        return self.value(field_name, True) is None

    def set_generated(self, field_name: str, status: bool):
        """Mark a field as generated."""

        if status:
            if field_name not in self._generated_fields:
                self._generated_fields.append(field_name)
        else:
            if field_name in self._generated_fields:
                self._generated_fields.remove(field_name)

    def is_generated(self, field_name: str) -> bool:
        """Return if the field has marked as generated."""

        return field_name in self._generated_fields

    def is_valid(self) -> bool:
        """Return if buffer object is valid."""

        try:
            if not self._orm_obj:
                return False
            value = getattr(self._orm_obj, self._cursor.metadata().primaryKey())  # noqa: F841
        except sqlalchemy.orm.exc.ObjectDeletedError:  # type: ignore [attr-defined] # noqa: F821
            return False

        return True



"""
Application package for resources.

This package holds all functions and classes that are like side resources.
"""

from pineboolib.application.projectmodule import Project
from pineboolib.core import settings
from typing import Dict, List, Any

PROJECT = Project()

SERIALIZE_LIST: Dict[int, List[str]] = {}
FILE_CLASSES: Dict[str, str] = {}
ID_SESSION: str = ""

PINEBOO_VER = "0.99.91.7"

SHOW_CURSOR_EVENTS: bool = False  # Enable show pnsqlcursor actions debug.
SHOW_CONNECTION_EVENTS: bool = False  # Enable show debug when connection is closed.
SHOW_NESTED_WARNING: bool = False  # Enable show nested debug.
VIRTUAL_DB: bool = True  # Enable :memory: database on pytest.
LOG_SQL: bool = False  # Enable sqlalchemy logs.
USE_WEBSOCKET_CHANNEL: bool = False  # Enable websockets features.
USE_MISMATCHED_VIEWS: bool = False  # Enable mismatched views.
RECOVERING_CONNECTIONS: bool = False  # Recovering state.
AUTO_RELOAD_BAD_CONNECTIONS: bool = False  # Auto reload bad conecctions.
USE_REPORT_VIEWER: bool = False  # Enable internal report viewer.
ENABLE_ACLS: bool = True  # Enable acls usage.
USE_INTERACTIVE_GUI: bool = True  # Enable interactiveGUI value.
ENABLE_CALL_EXCEPTIONS: bool = True  # Enable QSA calls exceptions.
PARSE_PROJECT_ON_INIT: bool = settings.CONFIG.value("ebcomportamiento/parseProject", False)
USE_ALTER_TABLE_LEGACY: bool = True
PERSISTENT: Dict[str, Any] = {}
USE_FLFILES_FOLDER_AS_STATIC_LOAD: bool = True
TESTING_MODE: bool = False  # True when testing_mode else False
ALLOW_ALTER_TABLE: bool = True
PROJECT_NAME: str = None  # Nombre del proyecto
EXTERNAL_FOLDER: str = None  # Carpeta externa
UPDATE_FLFILES_FROM_FLFOLDER: bool = False  # Actualizar ficheros flfiles
FRAMEWORK_DEBUG_LEVEL: int = 20  # Nivel de debug 30 .... 0 en bloques de 5



# -*- coding: utf-8 -*-
"""Check the application dependencies."""

import sys
from pineboolib.core.utils import logging, utils_base, check_dependencies as core_check_dependencies
from pineboolib import application

LOGGER = logging.get_logger(__name__)


def check_dependencies(dict_: "core_check_dependencies.DependencyCheck", exit: bool = True) -> bool:
    """
    Check if a package is installed and return the result.

    @param dict_. Dict with the name of the agency and the module to be checked.
    @param exit . Exit if dependence fails.
    """
    dep_error: "core_check_dependencies.DependencyError" = (
        core_check_dependencies.get_dependency_errors(dict_)
    )
    if not dep_error:
        return True
    msg = ""
    LOGGER.debug("Error trying to import modules:\n%s", "\n\n".join(dep_error.values()))
    LOGGER.warning("Unmet dependences:")
    for (dep, suggestedpkg), errormsg in dep_error.items():
        LOGGER.warning("Install package %s for %s", suggestedpkg, dep)
        msg += "Instale el paquete %s.\n%s" % (suggestedpkg, errormsg)
        if dep == "pyfpdf":
            msg += "\n\n\n Use pip3 install -i https://test.pypi.org/simple/ pyfpdf==1.7.3"

    if exit:
        if application.PROJECT.DGI.useDesktop() and application.PROJECT.DGI.localDesktop():
            from pineboolib.q3widgets.messagebox import MessageBox

            MessageBox.warning(None, "Pineboo - Dependencias Incumplidas -", msg, MessageBox.Ok)

        if not utils_base.is_deployed():
            sys.exit(32)

    return False



"""
Tests for Orm on qsa.
"""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa
from pineboolib.application.database.orm.utils import do_flush


class TestOrm(unittest.TestCase):
    """Test Orm."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_load(self) -> None:
        """Load model."""
        qsa.thread_session_new()
        class1_ = qsa.from_project("flareas_orm")
        obj_ = qsa.orm_("flareas")()
        self.assertEqual(class1_, obj_.__class__)
        qsa.thread_session_free()

    # ===============================================================================
    #     def test_sessions_isolation(self) -> None:
    #
    #         session1_ = qsa.session()
    #         session2_ = qsa.session("dbAux")
    #         session3_ = qsa.session("aux")
    #
    #         self.assertNotEqual(session1_, session2_)
    #         self.assertNotEqual(session1_, session3_)
    #         self.assertNotEqual(session2_, session3_)
    # ===============================================================================

    def test_create_object(self) -> None:
        """Create object."""

        class_ = qsa.from_project("flareas_orm")
        self.assertTrue(class_)

        obj_ = class_()

        setattr(obj_, "bloqueo", True)
        setattr(obj_, "idarea", "A")
        setattr(obj_, "descripcion", "Area A")

        self.assertEqual(obj_.idarea, "A")
        self.assertEqual(obj_.bloqueo, True)
        self.assertEqual(getattr(obj_, "descripcion", ""), "Area A")

    def test_insert_to_database(self) -> None:
        """Insert object to database."""

        session_ = qsa.session()
        class_ = qsa.from_project("flareas_orm")

        obj_ = class_()

        setattr(obj_, "bloqueo", True)
        setattr(obj_, "idarea", "A")
        setattr(obj_, "descripcion", "Area A")
        session_.begin()
        session_.add(
            obj_
        )  # Introduce el nuevo registro en la BD. A partir de ahora los cambios posteriores se guardarán en la BD.
        # res_1 = session_.execute("SELECT idarea FROM flareas WHERE idarea = 'A'")
        # self.assertFalse(res_1.returns_rows)
        do_flush(session_, [obj_])  # Aplica el cambio en la BD.
        res_2 = session_.execute("SELECT idarea FROM flareas WHERE idarea = 'A'")  # type: ignore [arg-type]
        self.assertTrue(res_2.returns_rows)  # type: ignore [attr-defined]

        obj2_ = session_.query(class_).get("A")  # Recupera el registro de la BD

        self.assertEqual(obj_, obj2_)
        session_.rollback()

    def test_delete_from_database(self) -> None:
        """Insert object to database."""

        session_ = qsa.session()
        class_ = qsa.from_project("flareas_orm")

        obj_ = class_()

        setattr(obj_, "bloqueo", True)
        setattr(obj_, "idarea", "A")
        setattr(obj_, "descripcion", "Area A")

        session_.begin()
        session_.add(obj_)
        session_.commit()  # Se cierra la sesión (Transacción)

        session_2 = qsa.session()
        obj2_ = session_2.query(class_).get("A")  # Recupera el registro de la BD
        self.assertTrue(obj2_)
        session_2.begin()
        session_2.delete(obj2_)
        session_2.commit()

        session_3 = qsa.session()
        obj3_ = session_3.query(class_).get("A")  # Recupera el registro de la BD
        self.assertFalse(obj3_)

    def test_modify_data(self) -> None:
        """Insert object to database."""

        session_ = qsa.thread_session_new()
        class_ = qsa.from_project("flareas_orm")

        obj_ = class_()

        setattr(obj_, "bloqueo", True)
        setattr(obj_, "idarea", "B")
        setattr(obj_, "descripcion", "Area B")
        session_.begin()
        self.assertTrue(obj_.save())
        # session_.add(obj_)  # Introduce el nuevo registro en la BD
        session_.commit()
        session_2 = qsa.session()
        session_2.begin()
        obj2_ = session_2.query(class_).get("B")  # Recupera el registro de la BD
        self.assertEqual(obj2_.descripcion, "Area B")  # type: ignore [union-attr]
        obj2_.descripcion = "Area B modificada"  # type: ignore [union-attr]
        session_2.commit()  # Guarda el cambio permanentemente.

        session_3 = qsa.session()
        obj3_ = session_3.query(class_).get("B")
        self.assertEqual(obj3_.descripcion, "Area B modificada")  # type: ignore [union-attr]
        qsa.thread_session_free()

    def test_legacy_metadata(self) -> None:
        """Compares metadata with rom metadata."""

        aq_app = qsa.aqApp

        class_ = qsa.from_project("flareas_orm")

        metadata = aq_app.db().manager().metadata("flareas")
        self.assertTrue(metadata)
        if metadata is not None:
            self.assertEqual(metadata.name(), class_.legacy_metadata["name"])
            self.assertEqual(metadata.alias(), class_.legacy_metadata["alias"])

            result = None
            for field in class_.legacy_metadata["fields"]:
                if field["name"] == "bloqueo":
                    result = field["default"]
                    break

            self.assertEqual(
                metadata.field("bloqueo").defaultValue(), result  # type: ignore [union-attr]
            )

    def test_save_points(self) -> None:
        """Save points."""

        session_ = (
            qsa.thread_session_new()
        )  # implica nueva Transaccion si en la llama anterior se hizo rollback o commit.
        # Si no continua en la transaccion que se abrio la última vez

        class_ = qsa.from_project("flareas_orm")
        obj_ = class_()
        obj_.idarea = "C"
        obj_.descripcion = "Descripción C"
        obj_.bloqueo = True
        session_.begin()
        session_.add(obj_)

        session_.begin_nested()  # Save point

        obj_.descripcion = "Descripción Nueva"

        obj2_ = session_.query(class_).get("C")

        self.assertEqual(obj_.descripcion, "Descripción Nueva")
        self.assertEqual(obj2_.descripcion, "Descripción Nueva")  # type: ignore [union-attr]

        session_.rollback()  # rollback save_point

        self.assertEqual(obj_.descripcion, "Descripción C")

        session_.rollback()  # rollback transaccion
        qsa.thread_session_free()

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test_messagebox module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.q3widgets import messagebox


class TestMessageBox(unittest.TestCase):
    """TestMessageBox Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_buttons(self) -> None:
        """Test if the buttons exists"""
        msg_box = messagebox.MessageBox()
        self.assertTrue(hasattr(msg_box, "Yes"))
        self.assertTrue(hasattr(msg_box, "No"))
        self.assertTrue(hasattr(msg_box, "NoButton"))
        self.assertTrue(hasattr(msg_box, "Ok"))
        self.assertTrue(hasattr(msg_box, "Cancel"))
        self.assertTrue(hasattr(msg_box, "Ignore"))

    def test_types(self) -> None:
        """Test types."""

        msg_question = messagebox.MessageBox.question("pruebaq")
        self.assertTrue(msg_question is not None)
        msg_warning = messagebox.MessageBox.warning("pruebaw")
        self.assertTrue(msg_warning is not None)
        msg_info = messagebox.MessageBox.information("pruebai")
        self.assertTrue(msg_info is not None)
        msg_crit = messagebox.MessageBox.critical("pruebac")
        self.assertTrue(msg_crit is not None)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Flreloadbatch module."""
# -*- coding: utf-8 -*-
from pineboolib.qsa import qsa


class FormInternalObj(qsa.FormDBWidget):
    """FormInternalObj class."""

    def init(self) -> None:
        """Init function."""
        pass

    def main(self) -> None:
        """Entry function."""
        util = qsa.FLUtil()
        setting = "scripts/sys/modLastDirModules_%s" % qsa.sys.nameBD()
        last_dir = util.readSettingEntry(setting)
        modules_dir = qsa.FileDialog.getExistingDirectory(
            last_dir, util.translate("scripts", "Directorio de Módulos")
        )

        if not modules_dir:
            return
        qsa.Dir().setCurrent(modules_dir)

        command_result = qsa.Array()
        if util.getOS() == "WIN32":
            command_result = self.command_exec("cmd.exe /C dir /B /S *.mod")
        else:
            command_result = self.command_exec("find . -name *.mod")

        if not command_result.ok:
            qsa.MessageBox.warning(
                util.translate("scripts", "Error al buscar los módulos en el directorio:\n")
                + modules_dir,
                qsa.MessageBox.Ok,
                qsa.MessageBox.NoButton,
                qsa.MessageBox.NoButton,
            )
            return

        opciones = command_result.salida.split("\n")
        opciones.pop()
        modulos = self.options_chooser(opciones)
        if not modulos:
            return

        for modulo in modulos:
            qsa.sys.processEvents()
            if not self.load_module(modulo):
                qsa.MessageBox.warning(
                    util.translate("scripts", "Error al cargar el módulo:\n") + modulo,
                    qsa.MessageBox.Ok,
                    qsa.MessageBox.NoButton,
                    qsa.MessageBox.NoButton,
                )
                return

        util.writeSettingEntry(setting, modules_dir)
        app_ = qsa.aqApp
        if app_ is None:
            return

        app_.reinit()

    def command_exec(self, comando: str) -> qsa.Array:
        """Execute a command and return a value."""
        res = qsa.Array()
        qsa.ProcessStatic.execute(comando)
        if qsa.ProcessStatic.stderr != "":
            res["ok"] = False
            res["salida"] = qsa.ProcessStatic.stderr
        else:
            res["ok"] = True
            res["salida"] = qsa.ProcessStatic.stdout

        return res

    def load_module(self, nombre_fichero: str) -> bool:
        """Load a module and return True if loaded."""
        util = qsa.FLUtil()
        if util.getOS() == "WIN32":
            nombre_fichero = nombre_fichero[0 : len(nombre_fichero) - 1]

        return qsa.from_project("formflreloadlast").load_module(nombre_fichero)

    def version_compare(self, ver1: str, ver2: str) -> int:
        """Compare two versions and return the highest."""

        return qsa.from_project("formflreloadlast").version_compare(ver1, ver2)

    def options_chooser(self, opciones: qsa.Array) -> qsa.Array:
        """Show a choose option dialog and return selected values."""
        util = qsa.FLUtil()
        dialog = qsa.Dialog()
        dialog.okButtonText = util.translate("scripts", "Aceptar")
        dialog.cancelButtonText = util.translate("scripts", "Cancelar")
        bgroup = qsa.GroupBox()
        bgroup.setTitle(util.translate("scripts", "Seleccione módulos a cargar"))
        dialog.add(bgroup)
        resultado = qsa.Array()
        check_box_list = qsa.Array()
        for num, opcion in enumerate(opciones):
            check_box_list[num] = qsa.CheckBox()
            bgroup.add(check_box_list[num])
            check_box_list[num].text = opcion
            check_box_list[num].checked = True

        if dialog.exec_():
            for num, opcion in enumerate(opciones):
                if check_box_list[num].checked:
                    resultado[len(resultado)] = opciones[num]

        return resultado if len(resultado) else qsa.Array()



"""
Collect information from the query, such as field tables, lines, etc ...
"""

from pineboolib import application, logging
from pineboolib.application.database import utils

import datetime
import re
from typing import Optional, Dict, Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import ifieldmetadata  # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


class SqlInspector(object):
    """SqlInspector Class."""

    _sql_list: List[str]
    _sql: str
    _sql_original: str
    _invalid_tables: List[str]
    _mtd_fields: Dict[int, "ifieldmetadata.IFieldMetaData"]
    _field_list: Dict[str, int]
    _table_names: List[str]
    _alias: Dict[str, str]
    _posible_float: bool
    _list_sql: List[str]

    _suspected_injection: Optional[str]
    _suspicious_keywords: List[str] = [  # en minuscula!!
        "ascii(",
        "substring(",
        "select",
        "coalesce",
        "cast(",
        "character",
        "/**/",
        "current_database()",
        "::text",
        "sleep",
        "waitfor",
        "benchmark",
        "selectchar",
    ]
    _suspicious_characters: List[str] = ["(", ")", "'", '"', " ", ",", "/*", "*/"]
    _suspicious_bypass_union_time: List[str] = [
        "adminor1=1",
        "convertintunionall",
        "unionallselect",
        "unionselect@",  # unionselect@@version
        "selectsleep",
        "selectbecnhmark",
        "orderby1sleep",
        ";waitfordelay",
        "orpg_sleep",
        "fromselectsleep",
        "selectsubsctring@",
    ]

    def __init__(self) -> None:
        """
        Initialize the class.
        """
        self._sql = ""
        self._sql_original = ""
        self._list_sql = []
        self._field_list = {}
        self._table_names = []
        self._mtd_fields = {}
        self._invalid_tables = []
        self._suspected_injection = None
        # self.set_sql(sql_text)
        # self.resolve()

        # self.table_names()
        # self.field_names()

    def resolve(self) -> None:
        """Resolve query."""
        self._invalid_tables = []
        self._mtd_fields = {}
        self._field_list = {}
        self._table_names = []
        self._alias = {}
        self._list_sql = []
        self._posible_float = False
        if self._sql.startswith("show"):
            return

        self._resolve_fields()

    def mtd_fields(self) -> Dict[int, "ifieldmetadata.IFieldMetaData"]:
        """
        Return a dictionary with the fields of the query.

        @return fields dictionary.
        """

        return self._mtd_fields

    def get_from(self) -> str:
        """Return from clausule."""
        ret_ = ""
        if "from" in self._list_sql:
            index_from = self._list_sql.index("from")
            if "where" in self._list_sql:
                index_where = self._list_sql.index("where")
                ret_ = " ".join(self._list_sql[index_from + 1 : index_where])
            else:
                ret_ = " ".join(self._list_sql[index_from + 1 :])

        return ret_

    def get_where(self) -> str:
        """Return where clausule."""

        ret_ = ""
        if "where" in self._list_sql:
            index_where = self._list_sql.index("where")
            if "group" in self._list_sql:
                ret_ = " ".join(self._list_sql[index_where + 1 : self._list_sql.index("group")])
            elif "order" in self._list_sql:
                ret_ = " ".join(self._list_sql[index_where + 1 : self._list_sql.index("order")])
            else:
                ret_ = " ".join(self._list_sql[index_where + 1 :])

        return ret_

    def get_order_by(self) -> str:
        """Return order by clausule."""
        ret_ = ""
        if "order" in self._list_sql:
            index_order = self._list_sql.index("order")
            ret_ = " ".join(self._list_sql[index_order + 2 :])

        return ret_

    def table_names(self) -> List[str]:
        """
        Return a list with the tables of the query.

        @return tables list.
        """

        return self._table_names

    def set_table_names(self, table_names: List[str]) -> None:
        """
        Set a list with the tables of the query.

        @return tables list.
        """

        self._table_names = table_names

    def sql(self) -> str:
        """
        Return sql string.
        """

        return self._sql

    def set_sql(self, sql: str) -> None:
        """Set sql query."""
        self._sql_original = sql
        sql = sql.lower()
        sql = sql.replace("\n", " ")
        sql = sql.replace("\t", " ")
        sql = sql.strip()

        self._sql = sql

    def field_names(self) -> List[str]:
        """
        Return a list with the name of the fields.

        @return fields list.
        """
        return [k for k, v in sorted(self._field_list.items(), key=lambda item: item[1])]

    def field_list(self) -> Dict[str, int]:
        """
        Return a Dict with name and position.

        @return fields list.
        """

        return self._field_list

    def fieldNameToPos(self, name: str) -> int:
        """
        Return the position of a field, from the name.

        @param name. field name.
        @return index position.
        """
        name = name.strip()
        if name in self._field_list.keys():
            return self._field_list[name]
        else:
            if name.find(".") > -1:
                table_name = name[0 : name.find(".")]
                field_name = name[name.find(".") + 1 :]
                if table_name in self._alias.keys():
                    table_name = self._alias[table_name]

                    field_name = "%s.%s" % (table_name, field_name)
                    if field_name in self._field_list.keys():
                        return self._field_list[field_name]
                else:
                    # probando a cambiar tabla por alias
                    for alias in self._alias.keys():
                        if self._alias[alias] == table_name:
                            field_name = "%s.%s" % (alias, field_name)
                            if field_name in self._field_list.keys():
                                return self._field_list[field_name]

            else:
                for table_name in self.table_names():
                    field_name = "%s.%s" % (table_name, name)
                    if field_name in self._field_list.keys():
                        return self._field_list[field_name]

                    for alias in self._alias.keys():
                        if self._alias[alias] == table_name:
                            field_name = "%s.%s" % (alias, name)
                            if field_name in self._field_list.keys():
                                return self._field_list[field_name]

        raise Exception(
            "No se encuentra el campo %s en la query:\n%s.\ncampos: %s"
            % (name, self._sql, self.field_list())
        )

    def posToFieldName(self, pos: int) -> str:
        """
        Return the name of a field, from the position.

        @param name. field name.
        @return field name.
        """
        for k in self._field_list.keys():
            if int(self._field_list[k]) == pos:
                idx_inicio = self._sql.find(k)
                return self._sql_original[idx_inicio : idx_inicio + len(k)]

        raise Exception("fieldName not found! %s")

    def _resolve_fields(self) -> None:
        """
        Break the query into the different data.
        """
        self._sql = self._sql.replace(" cast(", " cast (")
        list_sql = self._sql.split(" ")
        self._list_sql = list_sql
        if list_sql[0] == "select":
            if "from" not in list_sql:
                return  # Se entiende que es una consulta especial

            index_from = list_sql.index("from")
            new_fields_list: List[str] = []
            fields_list = list_sql[1:index_from]
            for field in fields_list:
                field = field.replace(" ", "")
                if field.find(",") > -1:
                    extra_fields: List[str] = field.split(",")
                    new_fields_list = new_fields_list + extra_fields
                else:
                    new_fields_list.append(field)

            fields_list = new_fields_list
            new_fields_list = []
            inicio_parentesis: List[str] = []
            composed_field: Dict[str, List[str]] = {}
            for field in list(fields_list):
                # Comprueba si hay field_names compuestos
                if (field.find("(") > -1 and not field.find(")") > -1) or field in [
                    "case",
                    "cast",
                ]:  # si es multiple de verdad
                    # Contamos los parentesis
                    if field in ["case", "cast"]:
                        inicio_parentesis.append(str(len(inicio_parentesis) + 1))
                        composed_field[inicio_parentesis[-1]] = []

                    else:
                        segmento = field[field.find("(") :]
                        while segmento.find("(") > -1 and not segmento.find(")") > -1:
                            inicio_parentesis.append(str(len(inicio_parentesis) + 1))
                            composed_field[inicio_parentesis[-1]] = []
                            try:
                                segmento = segmento[segmento.find("(") + 1 :]
                            except Exception:
                                break

                    composed_field[inicio_parentesis[-1]].append(field)
                    continue
                elif field == "":
                    continue
                elif (
                    (field.find(")") > -1 and not field.find("(") > -1)
                    or field == "end"
                    and inicio_parentesis
                ):  # si es multiple de verdad
                    if field == "end":
                        composed_field[inicio_parentesis[-1]].append(field)
                        if len(inicio_parentesis) == 1:
                            new_fields_list.append(" ".join(composed_field[inicio_parentesis[-1]]))
                        else:
                            composed_field[inicio_parentesis[-2]] += composed_field[
                                inicio_parentesis[-1]
                            ]

                        composed_field[inicio_parentesis[-1]] = []
                        del composed_field[inicio_parentesis[-1]]
                        del inicio_parentesis[-1]

                    else:
                        segmento = field[field.find(")") :]
                        composed_field[inicio_parentesis[-1]].append(field)
                        while segmento.find(")") > -1 and not field.find("(") > -1:
                            if len(inicio_parentesis) == 1:
                                new_fields_list.append(
                                    " ".join(composed_field[inicio_parentesis[-1]])
                                )
                            else:
                                composed_field[inicio_parentesis[-2]] += composed_field[
                                    inicio_parentesis[-1]
                                ]
                            composed_field[inicio_parentesis[-1]] = []
                            del composed_field[inicio_parentesis[-1]]
                            del inicio_parentesis[-1]
                            try:
                                segmento = segmento[segmento.find(")") + 1 :]
                            except Exception:
                                break

                elif inicio_parentesis:  # si estoy en medio de un multiple
                    composed_field[inicio_parentesis[-1]].append(field)
                else:
                    new_fields_list.append(field)

            # Repasa si hay alias en los fieldnames.
            old_list = list(new_fields_list)
            new_fields_list = []
            expect_alias = False
            for pos, field_name in enumerate(old_list):
                num = len(new_fields_list)

                if pos > 1 and old_list[pos - 1] == "on" and old_list[pos - 2] == "distinct":
                    new_fields_list = new_fields_list[: num - 2]
                    continue

                if field_name == "as":
                    expect_alias = True
                    continue
                else:
                    if expect_alias:
                        expect_alias = False
                        del new_fields_list[num - 1]

                    new_fields_list.append(field_name)

            tables_list: List[str] = []
            if "where" in list_sql:
                index_where = list_sql.index("where")
                tables_list = list_sql[index_from + 1 : index_where]
            elif "order" in list_sql:
                index_order_by = list_sql.index("order")
                tables_list = list_sql[index_from + 1 : index_order_by]
            else:
                tables_list = list_sql[index_from + 1 :]

            if "group" in tables_list:
                index_group_by = tables_list.index("group")
                tables_list = tables_list[:index_group_by]

            tablas: List[str] = []
            self._alias = {}
            jump = 0
            # next_is_alias = None
            prev_ = ""
            last_was_table = False
            and_ = False
            for table in tables_list:
                if table == "cast":
                    jump += 3
                    last_was_table = False

                if jump > 0:
                    if and_ and jump == 2 and len(table) > 2:  # entra si = <= > is ...
                        jump = 0
                        and_ = False
                        last_was_table = False
                    else:
                        jump -= 1
                        prev_ = table
                        last_was_table = False
                        continue

                if table.find(")") > -1:
                    last_was_table = False
                    continue

                elif table == "on":
                    jump = 3
                    prev_ = table
                    last_was_table = False
                    continue

                elif table in ("left", "join", "right", "inner", "outer"):
                    prev_ = table
                    last_was_table = False
                    continue

                elif table in ("-", "+"):
                    last_was_table = False
                    continue

                elif table == "interval":
                    jump = 1
                    last_was_table = False
                    continue

                elif table == "as":
                    last_was_table = True
                    continue

                elif table in ("and", "or"):
                    jump = 3
                    and_ = table == "and"
                    last_was_table = False

                else:
                    if last_was_table:
                        self._alias[table] = prev_
                        last_was_table = False
                    else:
                        if table != "":
                            last_was_table = True
                            if table.endswith(","):
                                table = table[:-1]
                                last_was_table = False
                            if table not in tablas:
                                tablas.append(table)

                    prev_ = table

            temp_tl: List[str] = []
            for item in tablas:
                temp_tl = temp_tl + item.split(",")
            tablas = temp_tl

            fl_finish = []
            for field_name in new_fields_list:
                if field_name.find(".") > -1:
                    table_ = field_name[0 : field_name.find(".")]
                    field_ = field_name[field_name.find(".") + 1 :]

                    if field_ == "*":
                        mtd_table = application.PROJECT.conn_manager.manager().metadata(table_)
                        if mtd_table is not None:
                            for item in mtd_table.fieldListArray():
                                fl_finish.append(item)

                            continue

                #    if a_.find("(") > -1:
                #        a = a_[a_.find("(") + 1 :]
                #    else:
                #        a = a_

                # if a in self._alias.keys():
                #    field_name = "%s.%s" % (a_.replace(a, self._alias[a]), f_)

                fl_finish.append(field_name)

            self._create_mtd_fields(fl_finish, tablas)
            self._check_sql_injection(list_sql)  # Pasamos el where

    def resolve_empty_value(self, pos: int) -> Any:
        """
        Return a data type according to field type and value None.

        @param pos. index postion.
        """

        if not self.mtd_fields():
            if self._sql.find("sum(") > -1:
                return 0
            return None

        type_ = "double"
        if pos not in self._mtd_fields.keys():
            if pos not in self._field_list.values():
                LOGGER.warning(
                    "SQL_TOOLS : resolve_empty_value : No se encuentra la posición %s", pos
                )
                return None
        else:
            mtd = self._mtd_fields[pos]
            if mtd is not None:
                type_ = mtd.type()

        return utils.resolve_empty_qsa_value(type_)

    def resolve_value(self, pos: int, value: Any, raw: bool = False) -> Any:
        """
        Return a data type according to field type.

        @param pos. index postion.
        """

        ret_: Any = None
        type_ = "double"
        field_metadata = None
        if not self.mtd_fields():
            if isinstance(value, datetime.timedelta):
                type_ = "timestamp"
                ret_ = utils.resolve_qsa_value(type_, value)
            elif isinstance(value, datetime.time):
                type_ = "time"
                ret_ = utils.resolve_qsa_value(type_, value)
            elif isinstance(value, datetime.date):
                type_ = "date"
                ret_ = utils.resolve_qsa_value(type_, value)
            else:
                return value
        else:
            if pos not in self._mtd_fields.keys():
                if pos not in self._field_list.values():
                    LOGGER.warning(
                        "SQL_TOOLS : resolve_value : No se encuentra la posición %s", pos
                    )
                    return None
            else:
                field_metadata = self._mtd_fields[pos]
                if field_metadata is not None:
                    type_ = field_metadata.type()

        if type_ == "pixmap":
            if application.PROJECT.conn_manager is None:
                raise Exception("Project is not connected yet")

            if field_metadata is None:
                raise Exception("Field metadata not found")

            table_metadata = field_metadata.metadata()
            if table_metadata is None:
                raise Exception("Metadata not found")
            if raw or not application.PROJECT.conn_manager.manager().isSystemTable(
                table_metadata.name()
            ):
                ret_ = application.PROJECT.conn_manager.manager().fetchLargeValue(value)
            else:
                ret_ = value
        elif ret_ is None:
            ret_ = utils.resolve_qsa_value(type_, value)

        return ret_

    def _create_mtd_fields(self, fields_list: list, tables_list: list) -> None:
        """
        Solve the fields that make up the query.

        @param fields_list. fields list.
        @param tables_list. tables list.
        """
        if application.PROJECT.conn_manager is None:
            raise Exception("Project is not connected yet")

        _filter = ["sum(", "max(", "distint("]

        self._mtd_fields = {}
        self._invalid_tables = []
        self._table_names = list(tables_list)
        # self._field_list = {k: n for n, k in enumerate(fields_list)}

        for number_, field_name_org in enumerate(list(fields_list)):
            self._field_list[field_name_org] = number_
            field_name = field_name_org
            for table_name in list(tables_list):
                mtd_table = application.PROJECT.conn_manager.manager().metadata(table_name)
                mtd_field = None
                if mtd_table is not None:
                    for fil in _filter:
                        if field_name.startswith(fil):
                            field_name = field_name.replace(fil, "")
                            field_name = field_name[:-1]

                    field_name_fixed = field_name
                    if field_name.find(".") > -1:
                        if table_name != field_name[0 : field_name.find(".")]:
                            continue
                        else:
                            field_name_fixed = field_name[field_name.find(".") + 1 :]
                    mtd_field = mtd_table.field(field_name_fixed)
                    if mtd_field is not None:
                        for existed_field in self._mtd_fields.values():
                            if existed_field.name() == mtd_field.name():
                                if (
                                    existed_field.metadata().name()  # type: ignore [union-attr]
                                    == mtd_field.metadata().name()  # type: ignore [union-attr]
                                ):
                                    LOGGER.info("%s already exists. Skipping" % mtd_field.name())
                                    continue
                        self._mtd_fields[number_] = mtd_field

                    # fields_list.remove(field_name_org)
                else:
                    if table_name not in self._invalid_tables:
                        self._invalid_tables.append(table_name)
                    # tables_list.remove(table_name)

    def suspected_injection(self) -> bool:
        """Return if the query contains suspicion of sql injection."""

        return False if self._suspected_injection is None else True

    def _check_sql_injection(self, where: List[str]) -> None:
        """Examine the query for suspected sql injection."""
        infected = ""

        # 1 concatenado.
        for word in (item for item in where if len(item) > 30):
            word_lower = word.lower()
            if [
                suspicious_word
                for suspicious_word in self._suspicious_keywords
                if suspicious_word in word_lower
            ]:
                infected = word
                break
        # 2 bypass
        if not infected:
            rep = dict((re.escape(character), "") for character in self._suspicious_characters)
            pattern = re.compile("|".join(rep.keys()))
            split_where = str(
                pattern.sub(lambda m: rep[re.escape(m.group(0))], " ".join(where))
            ).lower()
            for bypass in self._suspicious_bypass_union_time:
                if bypass in split_where:
                    infected = "%s -> %s" % (bypass, " ".join(where))
                    break

        if infected:
            self._suspected_injection = infected



"""Aqformdb module."""

from pineboolib import application

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.fllegacy.flformdb import FLFormDB  # pragma: no cover


def aq_form_db(*args) -> "FLFormDB":
    """Return a FLFormDB instance."""

    if application.PROJECT.conn_manager is None:
        raise Exception("Project is not connected yet")

    ac_xml = application.PROJECT.actions[args[0]]
    ac_xml.load_master_form()
    if ac_xml._master_widget is None:
        raise Exception("mainform_widget is emtpy!")

    return ac_xml._master_widget.form



"""
Version number normalization library.
"""
import re
from typing import List, Optional, Tuple, Any

from pineboolib.core.utils import logging

SubVersionNumber = int
SubVersionAppendedText = str
SubVersionTuple = Tuple[SubVersionNumber, SubVersionAppendedText]


LOGGER = logging.get_logger(__name__)


class VersionNumber:
    """
    Create version objects from string that can be easily be compared together.
    """

    is_null: bool  # True if no version at all.
    raw_text: str  # Raw version text or empty string if null.
    text: str  # Version text or empty string if null.
    normalized: List["SubVersionTuple"]  # Normalized version

    def __init__(self, version_string: Optional[str], default: Optional[str] = None) -> None:
        """
        Create Version object.
        """
        if default is not None and version_string is None:
            version_string = default

        self.is_null = version_string is None

        if version_string is None:
            version_string = ""

        self.raw_text = version_string
        self.text = version_string.strip().lstrip("v")
        self.normalized = [] if self.is_null else self.normalize_complex(version_string)

    def __eq__(self, other_any: Any) -> bool:
        """Compare for equality."""
        other: "VersionNumber" = self.to_version_number(other_any)
        return self.normalized == other.normalized

    def __neq__(self, other_any: Any) -> bool:
        """Compare for unequality."""
        other: "VersionNumber" = self.to_version_number(other_any)
        return self.normalized != other.normalized

    def __gt__(self, other_any: Any) -> bool:
        """Compare for greater than."""
        other: "VersionNumber" = self.to_version_number(other_any)
        return self.normalized > other.normalized

    def __lt__(self, other_any: Any) -> bool:
        """Compare for less than."""
        other: "VersionNumber" = self.to_version_number(other_any)
        return self.normalized < other.normalized

    def __ge__(self, other_any: Any) -> bool:
        """Compare for greater than equal."""
        other: "VersionNumber" = self.to_version_number(other_any)
        return self.normalized >= other.normalized

    def __le__(self, other_any: Any) -> bool:
        """Compare for less than equal."""
        other: "VersionNumber" = self.to_version_number(other_any)
        return self.normalized <= other.normalized

    def __str__(self) -> str:
        """Get string representation of the version."""
        return self.text

    def __repr__(self) -> str:
        """Get python representation of the version."""
        return "<%s %r>" % (str(self.__class__.__name__), None if self.is_null else self.text)

    @classmethod
    def to_version_number(self, other_any: Any) -> "VersionNumber":
        """Convert any input to VersionNumber."""
        other: "VersionNumber"
        if isinstance(other_any, str):
            other = VersionNumber(other_any)
        elif isinstance(other_any, VersionNumber):
            other = other_any
        else:
            raise ValueError("Can't compare VersionNumber with %r" % type(other_any))
        return other

    @classmethod
    def normalize_complex(cls, raw_text: str) -> List["SubVersionTuple"]:
        """
        Normalize a complex version as 1.0.5b-ubuntu0 into something that can be used for comparison.
        """
        decomposed_str: List[str] = raw_text.strip().lstrip("v").split(".")
        normalized: List[SubVersionTuple] = []
        subver: str
        for subver in decomposed_str:
            subver = subver.strip()
            subver_appended_part: SubVersionAppendedText = subver.lstrip("0123456789")

            subver_number_part = (
                subver[: -len(subver_appended_part)] if subver_appended_part else subver
            )

            subver_number: SubVersionNumber = int(subver_number_part) if subver_number_part else -1
            normalized.append((subver_number, subver_appended_part))
        return normalized

    @classmethod
    def check(cls, mod_name: str, mod_ver: str, min_ver: str) -> bool:
        """Compare two version numbers and raise a warning if "minver" is not met."""
        if cls.normalize(mod_ver) < cls.normalize(min_ver):
            LOGGER.warning(
                "La version de <%s> es %s. La mínima recomendada es %s.", mod_name, mod_ver, min_ver
            )
            return False
        return True

    @classmethod
    def normalize(cls, value: str) -> List[int]:
        """Normalize version string numbers like 3.10.1 so they can be compared."""
        return [int(item) for item in re.sub(r"[^\d\.]", "", value).split(".")]



"""Flreportviewer module."""
from PyQt6 import QtWidgets, QtCore, QtXml

from pineboolib.core import decorators, settings
from pineboolib import application
from pineboolib.core.utils import utils_base
from pineboolib.application import qsatypes
from pineboolib.fllegacy import flsqlquery
from pineboolib.fllegacy import flsqlcursor
from pineboolib.fllegacy import flmanagermodules

from pdf2image import convert_from_path

from pineboolib.fllegacy.flreportengine import FLReportEngine
from pineboolib import logging

from typing import Any, List, Mapping, Sized, Union, Dict, Optional, Callable, TYPE_CHECKING
from PyQt6.QtGui import QPalette, QPixmap

from PIL.ImageQt import ImageQt

import shutil
import pathlib

if TYPE_CHECKING:
    from pineboolib.q3widgets import qmainwindow  # noqa: F401
    from PyQt6.QtGui import QImage  # noqa: F401


LOGGER = logging.get_logger(__name__)

AQ_USRHOME = "."  # FIXME


class FLReportViewer(QtWidgets.QWidget):
    """FLReportViewer class."""

    pdfFile: str
    Append: int
    Display: int
    PageBreak: int

    qry_: Any
    xml_data_: Any
    template_: Any
    _auto_close: bool
    slot_print_disabled: bool
    slot_exported_disabled: bool
    _style_name: str
    _report_engine: "FLReportEngine"

    PrintGrayScale = 0
    PrintColor = 1
    _w: QtWidgets.QWidget

    _report_engine: Optional[FLReportEngine]
    dpi_: int
    report_: List[Any]
    _num_copies: int
    _printer_name: str
    _color_mode: int  # 1 color, 0 gray_scale

    def __init__(
        self,
        parent: Optional[QtWidgets.QWidget] = None,
        name: Optional[str] = None,
        embed_in_parent: bool = False,
        report_engine: Optional["FLReportEngine"] = None,
    ) -> None:
        """Inicialize."""

        super().__init__(parent)

        # self.loop_ = False
        # self.eventloop = QtCore.QEventLoop()

        self.report_printed = False
        self.slot_print_disabled = False
        self.slot_exported_disabled = False
        self.printing_ = False
        self.embed_in_parent = True if parent and embed_in_parent else False
        self.ui_: Dict[str, QtCore.QObject] = {}

        self.Display = 1  # pylint: disable=invalid-name
        self.Append = 1  # pylint: disable=invalid-name
        self.PageBreak = 1  # pylint: disable=invalid-name
        self._report_engine = report_engine or FLReportEngine(self)
        self._w = None

        self.dpi_ = int(settings.SETTINGS.value("rptViewer/dpi", 300))
        self.report_ = []
        self._num_copies = 1
        self._printer_name = ""
        self._color_mode = 1
        self._page_count = -1

    def resolution(self) -> int:
        """Return resolution."""

        return self.dpi_

    def reportPages(self) -> List[Any]:
        """Return report pages."""
        return self.report_

    def setNumCopies(self, num_copies: int) -> None:
        """Set number of copies."""
        self._num_copies = num_copies

    def setPrinterName(self, name: str) -> None:
        """Set printer name."""
        self._printer_name = name

    def setColorMode(self, color_mode: int) -> None:
        """Set color mode."""

        self._color_mode = color_mode

    def slotFirstPage(self):
        """Positioning first page."""
        self._w.set_page(0)

    def slotLastPage(self):
        """Positioning last page."""
        cnt = len(self.report_)
        self._w.set_page(cnt - 1)

    def slotNextPage(self):
        """Positioning next page."""

        cnt = len(self.report_)
        current_page = self._w._current_page
        next_page = (current_page + 1) if current_page < cnt - 1 else current_page
        self._w.set_page(next_page)

    def slotPrevPage(self):
        """Positioning prev page."""

        current_page = self._w._current_page
        prev_page = (current_page - 1) if current_page > 0 else current_page
        self._w.set_page(prev_page)

    def slotPrintReport(self):
        """Print report."""
        if self.slot_print_disabled:
            return

        from PyQt6 import QtPrintSupport

        dialog = QtPrintSupport.QPrintDialog()
        if dialog.exec() == QtWidgets.QDialog.DialogCode.Accepted:
            self.report_printed = self._report_engine.printReport(dialog)
            if self._auto_close and self.report_printed:
                self._w.close()

    def slotZoomUp(self):
        """ZoomUp."""

        self._w._scale_factor += 0.5
        current_page = self._w._current_page
        self._w._current_page = -1
        self._w.set_page(current_page)

    def slotZoomDown(self):
        """ZoomDown."""

        if self._w._scale_factor <= 0:
            return

        self._w._scale_factor -= 0.5
        current_page = self._w._current_page
        self._w._current_page = -1
        self._w.set_page(current_page)

    @decorators.not_implemented_warn
    def exportFileCSVData(self):
        """exportFileCSVData."""

        if self.slot_exported_disabled:
            return

    def exportToPDF(self):
        """exportToPDF."""

        if self.slot_exported_disabled:
            return

        data = QtWidgets.QFileDialog.getSaveFileName(
            self, "Seleccione fichero", str(pathlib.Path.home()), "*.pdf"
        )
        file_name = data[0]
        if file_name:
            if not file_name.endswith(".pdf"):
                file_name += ".pdf"
            shutil.copy(self._report_engine._parser.get_file_name(), file_name)

    @decorators.not_implemented_warn
    def sendEMailPDF(self):
        """sendEMailPDF."""

    @decorators.not_implemented_warn
    def showInitCentralWidget(self, value: bool):
        """showInitCentralWidget."""

    @decorators.not_implemented_warn
    def saveSVGStyle(self):
        """saveSVGStyle."""

    @decorators.not_implemented_warn
    def saveSimpleSVGStyle(self):
        """saveSimpleSVGStyle."""

    @decorators.not_implemented_warn
    def loadSVGStyle(self):
        """loadSVGStyle."""

    def __getattr__(self, name: str) -> Callable:
        """Return attributes from report engine."""
        return getattr(self._report_engine, name, None)

    def report_engine(self) -> "FLReportEngine":
        """Return report engine."""

        if self._report_engine is None:
            raise Exception("_report_engine is not defined!")
        return self._report_engine

    def setReportEngine(self, report_engine: Optional[FLReportEngine] = None) -> None:
        """Set report engine."""

        if self._report_engine is report_engine:
            return

        self._report_engine = report_engine
        if self._report_engine is not None:
            self.template_ = self._report_engine.rptNameTemplate()
            self.qry_ = self._report_engine.rptQueryData()

    def exec_(self) -> str:
        """Show report."""
        # if self.loop_:
        #    print("FLReportViewer::exec(): Se ha detectado una llamada recursiva")
        #    return

        if self._report_engine and hasattr(self._report_engine, "_parser"):
            pdf_file = self._report_engine._parser.get_file_name()

        if not utils_base.is_library():
            if application.USE_REPORT_VIEWER:
                self._w = FLWidgetReportViewer(self)
                self._w.show()
                self._w._file_name = pdf_file
                self._w.refresh()
                self.slotFirstPage()
            else:
                qsatypes.sysbasetype.SysBaseType.openUrl(pdf_file)

        return pdf_file

    @decorators.beta_implementation
    def csvData(self) -> str:
        """Return csv data."""

        return self._report_engine.csvData() if self._report_engine else ""

    def renderReport(
        self,
        init_row: int = 0,
        init_col: int = 0,
        append_or_flags: Union[bool, Sized, Mapping[int, Any]] = None,
        display_report: bool = False,
    ) -> bool:
        """Render report."""
        if self._report_engine is None:
            LOGGER.warning("report_engine is mepty!")
            return False

        flags = [self.Append, self.Display]

        if isinstance(append_or_flags, bool):
            flags[0] = append_or_flags

            if display_report is not None:
                flags[0] = display_report
        elif isinstance(append_or_flags, list):
            if len(append_or_flags) > 0:
                flags[0] = append_or_flags[0]  # display
            if len(append_or_flags) > 1:
                flags[1] = append_or_flags[1]  # append
            if len(append_or_flags) > 2:
                flags.append(append_or_flags[2])  # page_break

        result = self._report_engine.renderReport(init_row, init_col, flags)
        self.report_ = self._report_engine._parser._document.pages
        return result

    def renderReport2(
        self,
        init_row: int = 0,
        init_col: int = 0,
        append_or_flags: Union[bool, Sized, Mapping[int, Any]] = None,
        display_report: bool = False,
    ) -> bool:
        """Render report."""

        return self.renderReport(init_row, init_col, append_or_flags, display_report)

    def setReportData(
        self, data: Union["flsqlcursor.FLSqlCursor", "flsqlquery.FLSqlQuery", "QtXml.QDomNode"]
    ) -> bool:
        """Set data to report."""
        if isinstance(data, flsqlquery.FLSqlQuery):
            self.qry_ = data
            if self._report_engine and self._report_engine.setReportData(data):
                self.xml_data_ = self._report_engine.rptXmlData()
                return True
            return False
        elif isinstance(data, flsqlcursor.FLSqlCursor):
            if not self._report_engine:
                return False
            return self._report_engine.setReportData(data)
        elif isinstance(data, QtXml.QDomNode):
            self.xml_data_ = data
            self.qry_ = None
            if not self._report_engine:
                return False
            return self._report_engine.setReportData(data)
        return False

    def setReportTemplate(
        self, template: Union["QtXml.QDomNode", str], style: Optional[str] = None
    ) -> bool:
        """Set template to report."""
        if isinstance(template, QtXml.QDomNode):
            self._xml_template = template
            self.template_ = ""

            if not self._report_engine:
                return False

            if style is not None:
                self.setStyleName(style)

            self._report_engine.setFLReportTemplate(template)

            return True
        else:
            self.template_ = template
            self._style_name = style
            if self._report_engine and self._report_engine.setFLReportTemplate(template):
                # self.setStyleName(style)
                self._xml_template = self._report_engine.rptXmlTemplate()
                return True

        return False

    @decorators.beta_implementation
    def sizeHint(self) -> QtCore.QSize:
        """Return sizeHint."""
        return self.sizeHint()

    @decorators.beta_implementation
    def reportPrinted(self) -> bool:
        """Return if report was printed."""
        return self.report_printed

    def disableSlotsPrintExports(self, disable_print: bool = False, disable_export: bool = False):
        """Disable export and print slots."""

        self.slot_print_disabled = disable_print
        self.slot_exported_disabled = disable_export

    def printReport(self) -> None:
        """Print a report."""

        printer_name = self._printer_name
        num_copies = self._num_copies
        color_mode = self._color_mode

        self.report_printed = self._report_engine.printReport(printer_name, num_copies, color_mode)

    def printReportToPDF(self, file_name: str = "") -> None:
        """Print report to pdf."""

        if self.slot_print_disabled:
            return

        pdf_file = self._report_engine._parser.get_file_name()
        if pdf_file and file_name:
            shutil.copyfile(pdf_file, file_name)
            self.report_printed = True

    @decorators.pyqt_slot(int)
    def setResolution(self, dpi: int) -> None:
        """Set resolution."""
        settings.SETTINGS.set_value("rptViewer/dpi", dpi)
        self.dpi_ = dpi
        current = self._w._current_page if self._w._current_page > -1 else 0
        self._w.refresh()
        self._w.set_page(current)

    @decorators.pyqt_slot(int)
    def setPixel(self, rel_dpi: int) -> None:
        """Set pixel size."""
        settings.SETTINGS.set_value("rptViewer/pixel", rel_dpi)
        self._report_engine._rel_dpi = rel_dpi

    def setDefaults(self) -> None:
        """Set default values."""

        if self._w is not None:
            self._w.setDefaults()

    @decorators.not_implemented_warn
    def updateReport(self) -> None:
        """Update report."""

        pass

    @decorators.not_implemented_warn
    def getCurrentPage(self) -> Any:
        """Return curent page."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return FLPicture(self.report_.getCurrentPage(), self)
        return 0

    @decorators.not_implemented_warn
    def getFirstPage(self) -> Any:
        """Return first page."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return FLPicture(self.report_.getFirstPage(), self)
        return 0

    @decorators.not_implemented_warn
    def getPreviousPage(self) -> Any:
        """Return previous page."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return FLPicture(self.report_.getPreviousPage(), self)
        return 0

    @decorators.not_implemented_warn
    def getNextPage(self) -> Any:
        """Return next page."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return FLPicture(self.report_.getNextPage(), self)
        return 0

    @decorators.not_implemented_warn
    def getLastPage(self) -> Any:
        """Return last page."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return FLPicture(self.report_.getLastPage(), self)
        return 0

    @decorators.not_implemented_warn
    def getPageAt(self, num: int) -> Any:
        """Return actual page."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return FLPicture(self.report_.getPageAt(i), self)
        return 0

    @decorators.not_implemented_warn
    def updateDisplay(self) -> None:
        """Update display."""
        self.slotUpdateDisplay()

    @decorators.not_implemented_warn
    def clearPages(self) -> None:
        """Clear report pages."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     self.report_.clear()
        pass

    @decorators.not_implemented_warn
    def appendPage(self) -> None:
        """Add a new page."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     self.report_.appendPage()
        pass

    @decorators.not_implemented_warn
    def getCurrentIndex(self) -> int:
        """Return current index position."""

        if self._w:
            return self._w._current_page or -1

        return -1

    @decorators.not_implemented_warn
    def setCurrentPage(self, idx: int) -> None:
        """Set current page index."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     self.report_.setCurrentPage(idx)
        pass

    @decorators.not_implemented_warn
    def setPageSize(
        self, size: Union[QtCore.QSize, int], orientation: Optional[int] = None
    ) -> None:
        """Set page size."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     self.report_.setPageSize(s)
        pass

    @decorators.not_implemented_warn
    def setPageOrientation(self, orientation: int) -> None:
        """Set page orientation."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     self.report_.setPageOrientation(o)
        pass

    @decorators.not_implemented_warn
    def setPageDimensions(self, dim: QtCore.QSize) -> None:
        """Set page dimensions."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     self.report_.setPageDimensions(dim)
        pass

    @decorators.not_implemented_warn
    def pageSize(self) -> QtCore.QSize:
        """Return page size."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return self.report_.pageSize()
        return -1

    @decorators.not_implemented_warn
    def pageOrientation(self) -> int:
        """Return page orientation."""
        # FIXME: self.report_ is just a List[]
        # if self.report_:
        #     return self.report_.pageOrientation()
        return -1

    def pageDimensions(self) -> QtCore.QSize:
        """Return page dimensions."""
        if self._report_engine and hasattr(self._report_engine, "_parser"):
            return self._report_engine._parser._page_size
        return -1

    def pageCount(self) -> int:
        """Return number of pages."""
        if self._report_engine:
            return self._report_engine.number_pages()
        return -1

    @decorators.not_implemented_warn
    def setStyleName(self, style: str) -> None:
        """Set style name."""
        # self._style_name = style

    # @decorators.beta_implementation
    # def setReportPages(self, pgs: Any) -> None:
    #    """Add pages to actual report."""
    #    self.setReportEngine(None)
    #    self.qry_ = None
    #    self.xml_data_ = QtXml.QDomNode()
    #    self.report_viewer.setReportPages(pgs.pageCollection() if pgs else 0)
    #    self.report_ = self.report_viewer.reportPages()

    @decorators.beta_implementation
    def setName(self, name: str) -> None:
        """Set report name."""
        self.name_ = name

    @decorators.beta_implementation
    def name(self) -> str:
        """Return report name."""
        return self.name_


class FLWidgetReportViewer(QtWidgets.QMainWindow):
    """FLWidgetReportViewer."""

    _internal: "FLReportViewer"
    _fr_mail: "QtWidgets.QFrame"
    _auto_close: bool
    _auto_widget: "QtWidgets.QCheckBox"
    _pages: List["QImage"]
    _file_name: str
    _image_label: "QtWidgets.QLabel"
    _scroll_area: "QtWidgets.QScrollArea"
    _scale_factor: int
    _current_page: int
    _default_pix: int = 780
    _default_res: int = 300

    def __init__(self, report_viewer: "FLReportViewer") -> None:
        """Initialize."""

        super().__init__()
        self.setObjectName("FLWidgetReportViewer")

        from pineboolib.q3widgets import qframe, qcheckbox, qspinbox

        self._internal = report_viewer
        self._internal._auto_close = settings.SETTINGS.value("rptViewer/autoClose", False)
        form_path = utils_base.filedir("fllegacy/forms/FLWidgetReportViewer.ui")
        self = flmanagermodules.FLManagerModules.createUI(form_path, None, self)
        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        self._fr_mail = self.findChild(qframe.QFrame, "frEMail")
        self._auto_widget = self.findChild(qcheckbox.QCheckBox, "chkAutoClose")
        self._pixel_control = self.findChild(qspinbox.QSpinBox, "spnPixel")
        self._resolution_control = self.findChild(qspinbox.QSpinBox, "spnResolution")

        self._pixel_control.setValue(self._internal._report_engine._rel_dpi)
        self._resolution_control.setValue(self._internal.dpi_)

        self._auto_widget.setChecked(self._auto_close)
        self._fr_mail.hide()
        self._image_label = QtWidgets.QLabel()
        self._image_label.setBackgroundRole(QPalette.ColorRole.Base)
        self._image_label.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Ignored
        )
        self._image_label.setScaledContents(True)
        self._scroll_area = QtWidgets.QScrollArea()
        self._scroll_area.setBackgroundRole(QPalette.ColorRole.Dark)
        self._scroll_area.setWidget(self._image_label)
        self._scroll_area.setVisible(False)

        self.setCentralWidget(self._scroll_area)
        self._scale_factor = 1
        self._current_page = -1

        self.hide()
        self._pages = []
        self._file_name = ""
        self.clear()

    def __getattr__(self, name: str) -> Any:
        """Return FLReportViewer attributes."""
        return getattr(self._internal, name, None)

    def setAutoClose(self) -> None:
        """Set autoclose."""

        self._internal._auto_close = self._auto_widget.isChecked()
        settings.SETTINGS.set_value("rptViewer/autoClose", self._internal._auto_close)

    def close(self) -> None:
        """Close form."""

        super().close()

    def refresh(self) -> None:
        """Set page to centralWidget."""
        self._pages = convert_from_path(
            self._file_name, dpi=self._internal.dpi_, output_folder=application.PROJECT.tmpdir
        )
        self._current_page = -1
        self.clear()

    def setDefaults(self) -> None:
        """Set defaults values."""

        self._pixel_control.setValue(self._default_pix)
        self._resolution_control.setValue(self._default_res)
        self._scale_factor = 1
        self.refresh()
        self.set_page(0)

    def set_page(self, num) -> None:
        """Set page."""
        if self._current_page == num:
            return
        image = self._pages[num] if self._pages else None
        if image is not None:
            self._image_label.clear()
            self._current_page = num
            image_qt = ImageQt(image)
            width_ = 780
            height_ = 591
            scaled_size = QtCore.QSize(width_ * self._scale_factor, height_ * self._scale_factor)
            scaled = image_qt.scaled(
                scaled_size,
                QtCore.Qt.AspectRatioMode.KeepAspectRatio,
                QtCore.Qt.TransformationMode.SmoothTransformation,
            )
            pix = QPixmap.fromImage(scaled)
            self._image_label.setPixmap(pix)
            self._image_label.adjustSize()
            self._scroll_area.setVisible(True)

    def clear(self) -> None:
        """Clear current page."""
        self._scroll_area.setVisible(False)
        self._current_page = None
        # print("Borra centralwidget")
